""" RabbitMQ Utility library for NuroPb
"""
import logging
from typing import Dict, Any, List, Optional
from urllib.parse import urlparse
from contextlib import contextmanager

import requests
import pika
from pika.channel import Channel

from nuropb.interface import PayloadDict, NuropbTransportError

logger = logging.getLogger(__name__)


def build_amqp_url(
    host: str, port: str | int, username: str, password: str, vhost: str
) -> str:
    """Creates an AMQP URL for connecting to RabbitMQ"""
    return f"amqp://{username}:{password}@{host}:{port}/{vhost}"


def build_rmq_api_url(
    scheme: str, host: str, port: str | int, username: str | None, password: str | None
) -> str:
    """Creates an HTTP URL for connecting to RabbitMQ management API"""
    if username is None or password is None:
        return f"{scheme}://{host}:{port}/api"
    return f"{scheme}://{username}:{password}@{host}:{port}/api"


def rmq_api_url_from_amqp_url(
    amqp_url: str, scheme: Optional[str] = None, port: Optional[int | str] = None
) -> str:
    """Creates an HTTP URL for connecting to RabbitMQ management API from an AMQP URL
    :param amqp_url: the AMQP URL to use
    :param scheme: the scheme to use, defaults to http
    :param port: the port to use, defaults to 15672
    :return: the RabbitMQ management API URL
    """
    url_parts = urlparse(amqp_url)
    username = url_parts.username
    password = url_parts.password
    host = url_parts.hostname if url_parts.hostname else "localhost"
    port = 15672 if port is None else port
    scheme = "http" if scheme is None else scheme
    return build_rmq_api_url(scheme, host, port, username, password)


def management_api_session_info(
    scheme: str,
    host: str,
    port: str | int,
    username: Optional[str] = None,
    password: Optional[str] = None,
    bearer_token: Optional[str] = None,
    verify: bool = False,
    **headers: Any,
) -> Dict[str, Any]:
    """Creates a requests session for connecting to RabbitMQ management API
    :param scheme: http or https
    :param host: the host name or ip address of the RabbitMQ server
    :param port: the port number of the RabbitMQ server
    :param username: the username to use for authentication
    :param password: the password to use for authentication
    :param bearer_token: the bearer token to use for authentication
    :param verify: whether to verify the SSL certificate
    :return: a requests session
    """
    api_url = build_rmq_api_url(scheme, host, port, username, password)
    if bearer_token:
        headers["Authorization"] = f"Bearer {bearer_token}"
    headers["Content-Type"] = "application/json"
    session = requests.Session()
    session.headers = headers
    session.verify = verify
    return {
        "api_url": api_url,
        "headers": headers,
    }


@contextmanager
def blocking_rabbitmq_channel(rmq_url: str) -> pika.channel.Channel:
    """Useful for initialisation of queues / exchanges."""
    connection = None
    try:
        parameters = pika.URLParameters(rmq_url)
        connection = pika.BlockingConnection(parameters)
        channel = connection.channel()
        yield channel
    except Exception as e:
        logger.error(
            "Error opening blocking connection to open rabbitmq channel: %s", e
        )
        raise
    finally:
        if connection is not None:
            connection.close()


def configure_nuropb_rmq(
    service_name: str,
    rmq_url: str,
    events_exchange: str,
    rpc_exchange: str,
    dl_exchange: str,
    dl_queue: str,
    service_queue: str,
    rpc_bindings: List[str],
    event_bindings: List[str],
    **kwargs: Any,
) -> bool:
    """Configure the RabbitMQ broker for this transport.

    Calls to this function are IDEMPOTENT. However, previously named exchanges, queues,
    and declared bindings are not be removed. These will have to be done manually as part
    of broker housekeeping. This is to prevent accidental removal of queues and exchanges.
    It is safe to call this function multiple times and while other services are running,
    as it will not re-declare exchanges, queues, or bindings that already exist.

    PRODUCTION AUTHORISATION NOTE: The RabbitMQ user used to connect to the broker must
    have the following permissions:
    - configure: .*
    - write: .*
    - read: .*
    - access to the vhost
    Client only applications and services should not have configuration permissions. For
    completeness, there may be specific implementation need to for a client only services
    to register service queue bindings, for example a client only service that is also a
    gateway or proxy service. In this case, the treating it as a service is the correct
    approach.

    Settings for Exchange and default dead letter configuration apply to all services that
    use the same RabbitMQ broker. The rpc and event bindings are exclusive to the service,
    and are not shared with other services.

    The response queues are not durable, and are auto-deleted when the connections close.
    This approach is taken as response queues are only used for RPC responses, and there
    is no need to keep and have to handle stale responses.

    There is experimental work underway using etcd to manage the runtime configuration of
    a service leader and service followers. This will allow for persistent response queues
    and other configuration settings to be shared across multiple instances of the same
    service. This is not yet ready for production use.
    Experimentation scope:
    - service leader election
    - named instances of a service each with their own persistent response queue
    - Notification and handling of dead letter messages relating to a service

    :param str service_name: The name of the service, being configured for
    :param str rmq_url: The URL of the RabbitMQ broker
    :param str events_exchange: The name of the events exchange
    :param str rpc_exchange: The name of the RPC exchange
    :param str dl_exchange: The name of the dead letter exchange
    :param str dl_queue: The name of the dead letter queue
    :param str service_queue: The name of the requests queue
    :param List[str] rpc_bindings: The list of RPC bindings
    :param List[str] event_bindings: The list of events bindings
    :param kwargs: Additional keyword argument overflow from the transport settings.
        - client_only: bool - True if this is a client only service, False otherwise
    :return: True if the RabbitMQ broker was configured successfully
    """
    if kwargs.get("client_only", False):
        logger.info("Client only service, not configuring RMQ")
        return True

    if len(rpc_bindings) == 0 or service_name not in rpc_bindings:
        rpc_bindings.append(service_name)

    with blocking_rabbitmq_channel(rmq_url) as channel:
        logger.info(f"Declaring the dead letter exchange: {dl_exchange}")
        """ Setting up dead letter handling - all requests are automatically sent to a dead letter queue.
        for all services. This is to ensure that no messages are lost, and can be inspected for debugging
        purposes. The dead letter queue is durable, and will survive a broker restart.
        """
        channel.exchange_declare(
            exchange=dl_exchange,
            exchange_type="fanout",
            durable=True,
        )
        logger.info(f"Declaring the dead letter queue: {dl_queue}")
        channel.queue_declare(queue=dl_queue)
        logger.info(
            f"Binding the dead letter queue: {dl_queue} to the dead letter exchange: {dl_exchange}"
        )
        channel.queue_bind(dl_queue, dl_exchange)

        logger.info(f"Declaring the events exchange: {events_exchange}")
        channel.exchange_declare(
            exchange=events_exchange,
            exchange_type="topic",
            durable=True,
        )
        logger.info(f"Declaring the rpc exchange: {rpc_exchange}")
        channel.exchange_declare(
            exchange=rpc_exchange,
            exchange_type="direct",
            durable=True,
        )
        logger.info(
            f"Declaring the request queue: {service_queue} to the rpc exchange: {rpc_exchange}"
        )
        requests_queue_config = {
            "durable": True,
            "auto_delete": False,
            "arguments": {"x-dead-letter-exchange": dl_exchange},
        }
        channel.queue_declare(queue=service_queue, **requests_queue_config)

        """ This NOTE is here for reference only. A response queue is not durable by default
        and is the responsibility of the client/service follower to declare on startup. 
        As the response queue is not durable, it is auto-deleted when the connection is closed. 
        This is because the response queue is only used for RPC responses, and we don't want to 
        keep stale responses around. See the notes above about etcd leader/follower configuration
        for more information.
        
        responses_queue_config = {"durable": False, "auto_delete": True}
        """

        """ Any new bindings will be registered here, however, previous bindings will not be removed.
            PLEASE TAKE NOTE, especially in the context of event bindings, this could lead to a lot of
            unnecessary traffic and debugging support issues.
        """
        for routing_key in rpc_bindings:
            logger.info("binding to {}".format(routing_key))
            channel.queue_bind(service_queue, rpc_exchange, routing_key)

        for routing_key in event_bindings:
            logger.info("binding to {}".format(routing_key))
            channel.queue_bind(service_queue, events_exchange, routing_key)

        rabbitmq_configured = True

    return rabbitmq_configured


def nack_message(
    channel: Channel,
    delivery_tag: int,
    properties: pika.spec.BasicProperties,
    mesg: PayloadDict | None,
    error: Exception | None = None,
) -> None:
    """nack_message: nack the message and requeue it, there was likely a recoverable problem with this instance
    while processing the message
    """
    if channel is None or not channel.is_open:
        raise NuropbTransportError(
            description="Unable to nack and requeue message, RMQ channel closed",
            payload=mesg,
            exception=error,
        )
    logger.warning(
        f"Nacking message, delivery_tag: {delivery_tag}, correlation_id: {properties.correlation_id}"
    )
    channel.basic_nack(delivery_tag=delivery_tag, requeue=True)


def reject_message(
    channel: Channel,
    delivery_tag: int,
    properties: pika.spec.BasicProperties,
    mesg: PayloadDict | None,
    error: Exception | None = None,
) -> None:
    """reject_message: If the message is not a request, then reject the message and move on"""
    if channel is None or not channel.is_open:
        raise NuropbTransportError(
            description="unable to reject message, RMQ channel closed",
            payload=mesg,
            exception=error,
        )
    logger.warning(
        f"Rejecting message, delivery_tag: {delivery_tag}, correlation_id: {properties.correlation_id}"
    )
    channel.basic_reject(delivery_tag=delivery_tag, requeue=False)


def ack_message(
    channel: Channel,
    delivery_tag: int,
    properties: pika.spec.BasicProperties,
    mesg: PayloadDict | None,
    error: Exception | None = None,
) -> None:
    """ack_message: ack the message"""
    if channel is None or not channel.is_open:
        raise NuropbTransportError(
            description="Unable to ack message, RMQ channel closed",
            payload=mesg,
            exception=error,
        )
    logger.warning(
        f"Acking message, delivery_tag: {delivery_tag}, correlation_id: {properties.correlation_id}"
    )
    channel.basic_ack(delivery_tag=delivery_tag)


def get_virtual_host_queues(api_url: str, vhost_url: str) -> Any | None:
    """Creates a virtual host on the RabbitMQ server using the REST API
    :param api_url: the url to the RabbitMQ API
    :param vhost_url: the virtual host to create

    :return: None
    """
    url_parts = urlparse(vhost_url)
    vhost = url_parts.path[1:] if url_parts.path.startswith("/") else url_parts.path
    api_url += f"/queues/{vhost}"
    headers: Dict[str, Any] = {}
    response = requests.get(
        url=api_url,
        headers=headers,
        verify=False,
    )
    if response.status_code == 404:
        return None
    else:
        response.raise_for_status()
        return response.json()


def get_virtual_hosts(api_url: str, vhost_url: str) -> Any | None:
    """Creates a virtual host on the RabbitMQ server using the REST API
    :param api_url: the url to the RabbitMQ API
    :param vhost_url: the virtual host to create

    :return: None
    """
    _ = vhost_url
    api_url += "/vhosts"
    headers: Dict[str, Any] = {}
    response = requests.get(
        url=api_url,
        headers=headers,
        verify=False,
    )
    if response.status_code == 404:
        return None
    else:
        response.raise_for_status()
        return response.json()


def create_virtual_host(api_url: str, vhost_url: str) -> None:
    """Creates a virtual host on the RabbitMQ server using the REST API
    :param api_url: the url to the RabbitMQ API
    :param vhost_url: the virtual host to create

    :return: None
    """
    url_parts = urlparse(vhost_url)
    vhost = url_parts.path[1:] if url_parts.path.startswith("/") else url_parts.path

    vhost_data = get_virtual_hosts(api_url, vhost_url)
    vhost_exists = False
    if vhost_data:
        vhost_exists = any([item["name"] == vhost for item in vhost_data])

    if vhost_exists:
        logger.info(f"vhost exists: {vhost}")
        return

    api_url += f"/vhosts/{vhost}"
    headers = {"content-type": "application/json"}
    data = {"configure": ".*", "write": ".*", "read": ".*"}

    response = requests.put(
        url=api_url,
        json=data,
        headers=headers,
        verify=False,
    )
    logger.info(f"vhost created: {vhost}")
    response.raise_for_status()


def delete_virtual_host(api_url: str, vhost_url: str) -> None:
    """Deletes a virtual host on the RabbitMQ server using the REST API
    :param api_url: the url to the RabbitMQ API
    :param vhost_url: the virtual host to delete

    :return: None
    """
    url_parts = urlparse(vhost_url)
    vhost = url_parts.path[1:] if url_parts.path.startswith("/") else url_parts.path

    vhost_data = get_virtual_hosts(api_url, vhost_url)
    vhost_exists = False
    if vhost_data:
        vhost_exists = any([item["name"] == vhost for item in vhost_data])

    if not vhost_exists:
        logger.info(f"vhost does not exist: {vhost}")
        return

    api_url += f"/vhosts/{vhost}"
    headers = {"content-type": "application/json"}
    response = requests.delete(
        url=api_url,
        headers=headers,
        verify=False,
    )
    logger.info(f"vhost deleted: {vhost}")
    response.raise_for_status()



import re


def obfuscate_credentials(url_with_credentials: str) -> str:
    """obfuscate_secret: obfuscate the password in the AMQP url
    :param url_with_credentials:
    :return: str
    """
    pattern = r"(:.*?@)"
    result = re.sub(
        pattern,
        lambda match: ":" + "x" * (len(match.group(0)) - 2) + "@",
        url_with_credentials,
    )
    return result



import logging
import functools
from typing import List, Set, Optional, Any, Dict, Awaitable, Literal, TypedDict
import asyncio
import time
from cryptography.hazmat.primitives.asymmetric import rsa

import pika
from pika import connection
from pika.adapters.asyncio_connection import AsyncioConnection
from pika.channel import Channel
from pika.exceptions import ChannelClosedByBroker, ProbableAccessDeniedError
import pika.spec
from pika.frame import Method

from nuropb.encodings.encryption import Encryptor
from nuropb.encodings.serializor import encode_payload, decode_payload
from nuropb.interface import (
    NuropbTransportError,
    TransportServicePayload,
    NUROPB_PROTOCOL_VERSIONS_SUPPORTED,
    NUROPB_MESSAGE_TYPES,
    TransportRespondPayload,
    MessageCallbackFunction,
    AcknowledgeAction,
    NUROPB_PROTOCOL_VERSION,
    NUROPB_VERSION,
    NuropbNotDeliveredError,
    NuropbCallAgainReject,
)
from nuropb.rmq_lib import (
    rmq_api_url_from_amqp_url,
    create_virtual_host,
    configure_nuropb_rmq,
)
from nuropb import service_handlers
from nuropb.service_handlers import (
    create_transport_response_from_rmq_decode_exception,
    error_dict_from_exception,
)
from nuropb.utils import obfuscate_credentials


class RabbitMQConfiguration(TypedDict):
    rpc_exchange: str
    events_exchange: str
    dl_exchange: str
    dl_queue: str
    service_queue: str
    response_queue: str
    rpc_bindings: List[str]
    event_bindings: List[str]
    default_ttl: int
    client_only: bool


logger = logging.getLogger(__name__)

connection.PRODUCT = "NuroPb Distributed RPC-Event Service Mesh Library"
""" TODO: configure the RMQ client connection attributes in the pika client properties.
See related TODO below in this module
"""

CONSUMER_CLOSED_WAIT_TIMEOUT = 10
""" The wait when shutting down consumers before closing the connection
"""

_verbose = False


@property
def verbose() -> bool:
    return _verbose


@verbose.setter
def verbose(value: bool) -> None:
    global _verbose
    _verbose = value
    service_handlers.verbose = value


""" Set to True to enable module verbose logging
"""


def decode_rmq_body(
    method: pika.spec.Basic.Deliver, properties: pika.spec.BasicProperties, body: bytes
) -> TransportServicePayload:
    """Map the incoming RabbitMQ message to python compatible dictionary as
    defined by ServicePayloadDict
    """
    _ = method  # Future placeholder
    service_message: TransportServicePayload = {
        "nuropb_protocol": properties.headers.get("nuropb_protocol"),
        "nuropb_type": properties.headers.get("nuropb_type"),
        "nuropb_payload": {},
        "correlation_id": properties.correlation_id,
        "trace_id": properties.headers.get("trace_id"),
        "ttl": properties.expiration,
    }
    if service_message["nuropb_protocol"] not in NUROPB_PROTOCOL_VERSIONS_SUPPORTED:
        raise ValueError(
            f"nuropb_protocol '{service_message['nuropb_protocol']}' is not supported"
        )
    if service_message["nuropb_type"] not in NUROPB_MESSAGE_TYPES:
        raise ValueError(
            f"message_type '{service_message['nuropb_type']}' is not supported"
        )
    service_message["nuropb_payload"] = decode_payload(body, "json")
    return service_message


class ServiceNotConfigured(Exception):
    """Raised when a service is not properly configured on the RabbitMQ broker.
    the leader will be expected to configure the Exchange and service queues
    """

    pass


class RMQTransport:
    """

    If RabbitMQ closes the connection, this class will stop and indicate
    that reconnection is necessary. You should look at the output, as
    there are limited reasons why the connection may be closed, which
    usually are tied to permission related issues or socket timeouts.

    If the channel is closed, it will indicate a problem with one of the
    commands that were issued and that should surface in the output as well.

    """

    _service_name: str
    _instance_id: str
    _amqp_url: str
    _rpc_exchange: str
    _events_exchange: str
    _dl_exchange: str
    _dl_queue: str
    _service_queue: str
    _response_queue: str
    _rpc_bindings: Set[str]
    _event_bindings: Set[str]
    _prefetch_count: int
    _default_ttl: int
    _client_only: bool
    _message_callback: MessageCallbackFunction
    _encryptor: Encryptor | None

    _connected_future: Any
    _disconnected_future: Any

    _is_leader: bool
    _is_rabbitmq_configured: bool

    _connection: AsyncioConnection | None
    _channel: Channel | None
    _consumer_tags: Set[Any]
    _consuming: bool
    _connecting: bool
    _closing: bool
    _connected: bool
    _was_consuming: bool

    def __init__(
        self,
        service_name: str,
        instance_id: str,
        amqp_url: str,
        message_callback: MessageCallbackFunction,
        rpc_exchange: Optional[str] = None,
        events_exchange: Optional[str] = None,
        dl_exchange: Optional[str] = None,
        dl_queue: Optional[str] = None,
        service_queue: Optional[str] = None,
        response_queue: Optional[str] = None,
        rpc_bindings: Optional[List[str] | Set[str]] = None,
        event_bindings: Optional[List[str] | Set[str]] = None,
        prefetch_count: Optional[int] = None,
        default_ttl: Optional[int] = None,
        client_only: Optional[bool] = None,
        encryptor: Optional[Encryptor] = None,
    ):  # NOSONAR
        """Create a new instance of the consumer class, passing in the AMQP
        URL used to connect to RabbitMQ.

        :param str service_name: The name of the service
        :param str instance_id: The instance id of the service
        :param str amqp_url: The AMQP url to connect with
        :param MessageCallbackFunction message_callback: The callback to call when a message is received
        :param str rpc_exchange: The name of the RPC exchange
        :param str events_exchange: The name of the events exchange
        :param str dl_exchange: The name of the dead letter exchange
        :param str dl_queue: The name of the dead letter queue
        :param str service_queue: The name of the requests queue
        :param str response_queue: The name of the responses queue
        :param List[str] rpc_bindings: The list of RPC bindings
        :param List[str] event_bindings: The list of events bindings
        :param int prefetch_count: The number of messages to prefetch defaults to 1, unlimited is 0.
                Experiment with larger values for higher throughput in your user case.
        :param int default_ttl: The default time to live for messages in milliseconds, defaults to 12 hours.
        :param bool client_only:
        :param Encryptor encryptor: The encryptor to use for encrypting and decrypting messages
        """
        self._connected = False
        self._closing = False
        self._connecting = False
        self._was_consuming = False
        self._consuming = False

        self._connection = None
        self._channel = None
        self._consumer_tags = set()

        self._client_only = False if client_only is None else client_only
        """ If client_only is True, then the transport will not configure a service queue with
        the handling of requests, commands and events disabled. 
        """
        if self._client_only:
            logger.info("Client only transport")
            rpc_bindings = []
            event_bindings = []

        self._encryptor = encryptor

        # Experiment with larger values for higher throughput.
        self._service_name = service_name
        self._instance_id = instance_id
        self._amqp_url = amqp_url
        self._rpc_exchange = rpc_exchange or "nuropb-rpc-exchange"
        self._events_exchange = events_exchange or "nuropb-events-exchange"
        self._dl_exchange = dl_exchange or "nuropb-dl-exchange"
        self._dl_queue = dl_queue or f"nuropb-{self._service_name}-dl"
        self._service_queue = service_queue or f"nuropb-{self._service_name}-service"
        self._response_queue = (
            response_queue
            or f"nuropb-{self._service_name}-{self._instance_id}-response"
        )
        self._rpc_bindings = set(rpc_bindings or [])
        self._event_bindings = set(event_bindings or [])
        self._prefetch_count = 1 if prefetch_count is None else prefetch_count
        self._default_ttl = default_ttl or 60 * 60 * 1000 * 12  # 12 hours
        self._message_callback = message_callback
        self._rpc_bindings.add(self._service_name)

        self._is_leader = True
        self._is_rabbitmq_configured = False

        self._connected_future = None
        self._disconnected_future = None

    @property
    def service_name(self) -> str:
        return self._service_name

    @property
    def instance_id(self) -> str:
        return self._instance_id

    @property
    def amqp_url(self) -> str:
        return self._amqp_url

    @property
    def is_leader(self) -> bool:
        return self._is_leader

    @is_leader.setter
    def is_leader(self, value: bool) -> None:
        """is_leader: set the transport's leader status"""
        self._is_leader = value

    @property
    def connected(self) -> bool:
        """connected: returns the connection status of the underlying transport
        :return: bool
        """
        return self._connected

    @property
    def rpc_exchange(self) -> str:
        """rpc_exchange: returns the name of the RPC exchange
        :return: str
        """
        return self._rpc_exchange

    @property
    def events_exchange(self) -> str:
        """events_exchange: returns the name of the events exchange
        :return: str
        """
        return self._events_exchange

    @property
    def response_queue(self) -> str:
        """response_queue: returns the name of the response queue
        :return: str
        """
        return self._response_queue

    @property
    def rmq_configuration(self) -> RabbitMQConfiguration:
        """rmq_configuration: returns the RabbitMQ configuration
        :return: Dict[str, Any]
        """
        return {
            "rpc_exchange": self._rpc_exchange,
            "events_exchange": self._events_exchange,
            "dl_exchange": self._dl_exchange,
            "dl_queue": self._dl_queue,
            "service_queue": self._service_queue,
            "response_queue": self._response_queue,
            "rpc_bindings": list(self._rpc_bindings),
            "event_bindings": list(self._event_bindings),
            "default_ttl": self._default_ttl,
            "client_only": self._client_only,
        }

    def configure_rabbitmq(
        self,
        rmq_configuration: Optional[RabbitMQConfiguration] = None,
        amqp_url: Optional[str] = None,
        rmq_api_url: Optional[str] = None,
    ) -> None:
        """configure_rabbitmq: configure the RabbitMQ transport with the provided configuration

        if rmq_configuration is None, then the transport will be configured with the configuration
        provided during the transport's __init__().

        if the virtual host in the build_amqp_url is not configured, then it will be created.

        :param rmq_configuration: RabbitMQConfiguration
        :param amqp_url: Optional[str] if not provided then self._amqp_url is used
        :param rmq_api_url: Optional[str] if not provided then it is created amqp_url
        :return: None
        """
        if rmq_configuration is None:
            rmq_configuration = self.rmq_configuration
        amqp_url = amqp_url or self._amqp_url
        if amqp_url is None:
            raise ValueError("amqp_url is not provided")
        rmq_api_url = rmq_api_url or rmq_api_url_from_amqp_url(amqp_url)
        if rmq_api_url is None:
            raise ValueError("rmq_api_url is not provided")
        try:
            create_virtual_host(rmq_api_url, amqp_url)
            configure_nuropb_rmq(
                service_name=self._service_name,
                rmq_url=amqp_url,
                events_exchange=rmq_configuration["events_exchange"],
                rpc_exchange=rmq_configuration["rpc_exchange"],
                dl_exchange=rmq_configuration["dl_exchange"],
                dl_queue=rmq_configuration["dl_queue"],
                service_queue=rmq_configuration["service_queue"],
                rpc_bindings=rmq_configuration["rpc_bindings"],
                event_bindings=rmq_configuration["event_bindings"],
            )
            self._is_rabbitmq_configured = True
        except Exception as err:
            logger.exception(
                f"Failed to configure RabbitMQ with the provided configuration:\n"
                f"Error: {err}\n"
                f" - amqp url: {obfuscate_credentials(amqp_url)}\n"
                f" - api url: {obfuscate_credentials(rmq_api_url)}\n"
                f" - RabbitMQ configuration: {rmq_configuration}\n"
            )
            raise err

    async def start(self) -> None:
        """Start the transport by connecting to RabbitMQ"""
        self._connected_future = self.connect()
        try:
            await self._connected_future
        except ProbableAccessDeniedError as err:
            vhost = self._amqp_url.split("/")[-1]
            if "ConnectionClosedByBroker: (530) 'NOT_ALLOWED - vhost" in str(err):
                raise ServiceNotConfigured(
                    f"The NuroPb configuration is missing from RabbitMQ for the virtual host: {vhost}"
                )
            else:
                logger.error(
                    f"Access denied to RabbitMQ for the virtual host {vhost}: {err}"
                )
                raise err
        except Exception as err:
            logger.error(f"Failed to connect to RabbitMQ: {err}")
            raise err

    async def stop(self) -> None:
        """Cleanly shutdown the connection to RabbitMQ by stopping the consumer
        with RabbitMQ. When RabbitMQ confirms the cancellation, on_cancelok
        will be invoked by pika, which will then closing the channel and
        connection. The IOLoop is started again because this method is invoked
        when CTRL-C is pressed raising a KeyboardInterrupt exception. This
        exception stops the IOLoop which needs to be running for pika to
        communicate with RabbitMQ. All commands issued prior to starting the
        IOLoop will be buffered but not processed.

        """
        if not self._closing:
            logger.info("Stopping")
            if self._consuming:
                await self.stop_consuming()
            self._disconnected_future = self.disconnect()
            await self._disconnected_future

    def connect(self) -> Awaitable[bool]:
        """This method initiates a connection to RabbitMQ, returning the connection handle.
        When the connection is established, the on_connection_open method
        will be invoked by pika.

        When the connection and channel is successfully opened, the incoming messages will
        automatically be handled by _handle_message()

        :rtype: asyncio.Future

        """
        if self._connected and not self._closing and self._connection is not None:
            raise RuntimeError("Already connected to RabbitMQ")
        if self._closing:
            raise RuntimeError("Can't open a RabbitMQ connection while it is closing")
        if self._connected_future is not None and not self._connected_future.done():
            raise RuntimeError("Already connecting to RabbitMQ")

        logger.info("Connecting to %s", obfuscate_credentials(self._amqp_url))

        self._connected_future = asyncio.Future()

        conn = AsyncioConnection(
            parameters=pika.URLParameters(self._amqp_url),
            on_open_callback=self.on_connection_open,
            on_open_error_callback=self.on_connection_open_error,
            on_close_callback=self.on_connection_closed,
        )
        self._connection = conn
        self._connecting = True
        return self._connected_future

    def disconnect(self) -> Awaitable[bool]:
        """This method closes the connection to RabbitMQ. the pika library events will drive
        the closing and reconnection process.
        :return: asyncio.Future
        """
        if self._connection is None:
            raise RuntimeError("RMQ transport is not connected")

        if self._connection.is_closing or self._connection.is_closed:
            raise RuntimeError("RMQ transport is already closing or closed")

        if (
            self._disconnected_future is not None
            and not self._disconnected_future.done()
        ):
            raise RuntimeError("Already closing to RabbitMQ")

        logger.info("Closing RMQ transport connection")
        self._disconnected_future = asyncio.Future()
        self._closing = True
        self._connection.close()
        return self._disconnected_future

    def on_connection_open(self, _connection: AsyncioConnection) -> None:
        """This method is called by pika once the connection to RabbitMQ has
        been established. It passes the handle to the connection object in
        case we need it, but in this case, we'll just mark it unused.

        :param pika.adapters.asyncio_connection.AsyncioConnection _connection:
           The connection
        """
        logger.info("Connection opened - now opening channel")

        self.open_channel()

    def on_connection_open_error(
        self, _connection: AsyncioConnection, err: Exception
    ) -> None:
        """This method is called by pika if the connection to RabbitMQ
        can't be established.

        :param pika.adapters.asyncio_connection.AsyncioConnection _connection:
           The connection
        :param Exception err: The error
        """
        logger.error("Connection open failed: %s", err)
        if self._connected_future is not None and not self._connected_future.done():
            self._connected_future.set_exception(err)

        if self._connecting:
            self._connecting = False
            # self.connect()

    def on_connection_closed(
        self, _connection: AsyncioConnection, reason: Exception
    ) -> None:
        """This method is invoked by pika when the connection to RabbitMQ is
        closed unexpectedly. Since it is unexpected, we will reconnect to
        RabbitMQ if it disconnects.

        :param pika.connection.Connection _connection: The closed connection obj
        :param Exception reason: exception representing reason for loss of
            connection.

        """
        logger.warning("Connection closed. reason: %s", reason)
        if (
            self._disconnected_future is not None
            and not self._disconnected_future.done()
        ):
            self._disconnected_future.set_result(True)

        self._connected = False
        self._channel = None

        if self._closing:
            self._closing = False
        else:
            self._connecting = False
            self.connect()

    def open_channel(self) -> None:
        """Open a new channel with RabbitMQ by issuing the Channel.Open RPC command. When RabbitMQ
        responds that the channel is open, the on_channel_open callback will be invoked by pika.
        """
        logger.info("Creating a new channel")
        if self._connection is None:
            raise RuntimeError("RMQ transport is not connected")
        self._connection.channel(on_open_callback=self.on_channel_open)

    def on_channel_open(self, channel: Channel) -> None:
        """This method is invoked by pika when the channel has been opened. The channel object is passed
         in so that we can make use of it.

        :param pika.channel.Channel channel: The channel object
        """
        logger.info("Channel opened")
        self._channel = channel
        self._channel.add_on_close_callback(self.on_channel_closed)
        self._channel.add_on_return_callback(self.on_message_returned)
        self.declare_response_queue()

    def on_channel_closed(self, channel: Channel, reason: Exception) -> None:
        """Invoked by pika when RabbitMQ unexpectedly closes the channel. Channels are usually closed
        if you attempt to do something that violates the protocol, such as re-declare an exchange or
        queue with different parameters. In this case, we'll close the connection to shut down the object.

        :param pika.channel.Channel channel: The closed channel
        :param Exception reason: why the channel was closed
        """
        if isinstance(reason, ChannelClosedByBroker):
            logger.critical(
                f"RabbitMQ channel {channel} closed by broker with reply_code: {reason.reply_code} "
                f"and reply_text: {reason.reply_text}"
            )
            if self._connected_future and not self._connected_future.done():
                self._connected_future.set_exception(Exception)
                self._connecting = False

        # investigate reasons and methods automatically reopen the channel.
        # until a solution is found it will be important to monitor for this condition

    def declare_service_queue(self, frame: pika.frame.Method) -> None:
        """Refresh the request queue on RabbitMQ by invoking the Queue.Declare RPC command. When it
        is complete, the on_service_queue_declareok method will be invoked by pika.

        This call is idempotent and will not fail if the queue already exists.
        """
        if not self._client_only:
            logger.info("Declaring request queue %s", self._service_queue)
            if self._channel is None:
                raise RuntimeError("RMQ transport channel is not open")

            cb = functools.partial(
                self.on_service_queue_declareok, _userdata=self._service_queue
            )
            service_queue_config = {
                "durable": True,
                "auto_delete": False,
                "arguments": {"x-dead-letter-exchange": self._dl_exchange},
            }
            self._channel.queue_declare(
                queue=self._service_queue, callback=cb, **service_queue_config
            )
        else:
            logger.info("Client only, not declaring request queue")
            # Check that passing None in here is OK
            self.on_bindok(frame, userdata=self._response_queue)

    def on_service_queue_declareok(
        self, frame: pika.frame.Method, _userdata: str
    ) -> None:
        logger.info(f"Refreshing rpc bindings for service queue: {self._service_queue}")
        if self._channel is None:
            raise RuntimeError("RMQ transport channel is not open")

        for routing_key in self._rpc_bindings:
            logger.info(
                f"Binding to"
                f" rpc exchange {self._rpc_exchange}"
                f" on routing key {routing_key}"
            )
            self._channel.queue_bind(
                self._service_queue, self._rpc_exchange, routing_key=routing_key
            )
        logger.info(
            f"Refreshing event bindings for service queue: {self._service_queue}"
        )
        for routing_key in self._event_bindings:
            logger.info(
                f"Binding to"
                f" events exchange {self._events_exchange}"
                f" on topic {routing_key}"
            )
            self._channel.queue_bind(
                self._service_queue, self._events_exchange, routing_key=routing_key
            )
        self.on_bindok(frame, userdata=self._response_queue)

    def declare_response_queue(self) -> None:
        """Set up the response queue on RabbitMQ by invoking the Queue.Declare RPC command. When it
        is complete, the on_response_queue_declareok method will be invoked by pika.
        """
        logger.info("Declaring response queue %s", self._response_queue)
        if self._channel is None:
            raise RuntimeError("RMQ transport channel is not open")

        cb = functools.partial(
            self.on_response_queue_declareok, _userdata=self._response_queue
        )
        response_queue_config = {"durable": False, "auto_delete": True}
        self._channel.queue_declare(
            queue=self._response_queue, callback=cb, **response_queue_config
        )

    def on_response_queue_declareok(
        self, frame: pika.frame.Method, _userdata: str
    ) -> None:
        """Method invoked by pika when the Queue.Declare RPC call made in setup_response_queue has
        completed. In this method we will bind request queue and the response queues. When this
        command is complete, the on_bindok method will be invoked by pika.

        No explicit binds required for the response queue as it relies on the default exchange and
        routing key to the name of the queue.

        :param pika.frame.Method frame: The Queue.DeclareOk frame
        :param str|unicode _userdata: Extra user data (queue name)
        """
        _ = frame
        logger.info("Response queue declared ok: %s", _userdata)
        self.declare_service_queue(frame)

    def on_bindok(self, _frame: pika.frame.Method, userdata: str) -> None:
        """Invoked by pika when the Queue.Bind method has completed. At this
        point we will set the prefetch count for the channel.

        :param pika.frame.Method _frame: The Queue.BindOk response frame
        :param str|unicode userdata: Extra user data (queue name)
        """
        logger.info("Response queue bound ok: %s", userdata)
        if self._channel is None:
            raise RuntimeError("RMQ transport channel is not open")

        self._channel.basic_qos(
            prefetch_count=self._prefetch_count, callback=self.on_basic_qos_ok
        )

    def on_basic_qos_ok(self, _frame: pika.frame.Method) -> None:
        """Invoked by pika when the Basic.QoS method has completed. At this
        point we will start consuming messages by calling start_consuming
        which will invoke the needed RPC commands to start the process.

        :param pika.frame.Method _frame: The Basic.QosOk response frame

        """
        logger.info("QOS set to: %d", self._prefetch_count)
        logger.info("Configure message consumption")

        """Add a callback that will be invoked if RabbitMQ cancels the consumer for some reason. 
        If RabbitMQ does cancel the consumer, on_consumer_cancelled will be invoked by pika.
        """
        if self._channel is None:
            raise RuntimeError("RMQ transport channel is not open")

        self._channel.add_on_cancel_callback(self.on_consumer_cancelled)

        # Start consuming the response queue these need their own handler as the ack type is automatic
        self._consumer_tags.add(
            self._channel.basic_consume(
                on_message_callback=functools.partial(
                    self.on_response_message, self._response_queue
                ),
                auto_ack=True,
                queue=self._response_queue,
                exclusive=True,
            )
        )

        # Start consuming the requests queue
        if not self._client_only:
            logger.info("Ready to consume requests, events and commands")
            self._consumer_tags.add(
                self._channel.basic_consume(
                    on_message_callback=functools.partial(
                        self.on_service_message, self._service_queue
                    ),
                    queue=self._service_queue,
                )
            )

        self._was_consuming = True
        self._consuming = True

        if self._connected_future:
            self._connected_future.set_result(True)
            self._connecting = False
        self._connected = True

    def on_consumer_cancelled(self, method_frame: pika.frame.Method) -> None:
        """Invoked by pika when RabbitMQ sends a Basic.Cancel for a consumer receiving messages.

        :param pika.frame.Method method_frame: The Basic.Cancel frame
        """
        logger.info(
            f"Consumer was cancelled remotely, closing the current channel. Reason: {method_frame}"
        )
        if self._channel:
            self._channel.close()

    def on_message_returned(
        self,
        channel: pika.channel.Channel,  # noqa
        method: pika.spec.Basic.Return,
        properties: pika.spec.BasicProperties,
        body: bytes,  # noqa
    ) -> None:
        """Called when message has been rejected by the server.

        callable callback: The function to call, having the signature callback(channel, method, properties, body)

        :param channel: pika.Channel
        :param method: pika.spec.Basic.Deliver
        :param properties: pika.spec.BasicProperties
        :param body: bytes
        """
        _ = channel, body
        correlation_id = properties.correlation_id
        trace_id = properties.headers.get("trace_id", "unknown")
        nuropb_type = properties.headers.get("nuropb_type", "unknown")
        nuropb_version = properties.headers.get("nuropb_version", "unknown")
        encrypted = properties.headers.get("encrypted", "") == "yes"

        logger.warning(
            f"Could not route {nuropb_type} message to service {method.routing_key} "
            f"correlation_id: {correlation_id} "
            f"trace_id: {trace_id} "
            f": {method.reply_code}, {method.reply_text}"
        )
        """ End the awaiting request future with a NuropbNotDeliveredError
        """
        if nuropb_type == "request":
            if encrypted and self._encryptor:
                body = self._encryptor.decrypt_payload(
                    payload=body,
                    correlation_id=correlation_id,
                )
            nuropb_payload = decode_payload(body, "json")
            request_method = nuropb_payload["method"]
            nuropb_payload["tag"] = nuropb_type
            nuropb_payload["error"] = {
                "error": "NuropbNotDeliveredError",
                "description": f"Service {method.routing_key} not available. unable to call method {request_method}",
            }
            nuropb_payload["result"] = None
            nuropb_payload.pop("service", None)
            nuropb_payload.pop("method", None)
            nuropb_payload.pop("params", None)
            nuropb_payload.pop("reply_to", None)
            message = TransportRespondPayload(
                nuropb_protocol=NUROPB_PROTOCOL_VERSION,
                correlation_id=correlation_id,
                trace_id=trace_id,
                ttl=None,
                nuropb_type="response",
                nuropb_payload=nuropb_payload,
            )
            metadata = {
                "start_time": time.time(),
                "service_name": self._service_name,
                "instance_id": self._instance_id,
                "is_leader": self._is_leader,
                "client_only": self._client_only,
                "nuropb_type": nuropb_type,
                "nuropb_version": nuropb_version,
                "correlation_id": properties.correlation_id,
                "trace_id": properties.headers.get("trace_id", "unknown"),
            }

            message_complete_callback = functools.partial(
                self.on_service_message_complete,
                channel,
                "",
                "",
                metadata,
            )
            self._message_callback(message, message_complete_callback, metadata)

        if verbose:
            raise NuropbNotDeliveredError(
                (
                    f"Could not route message {nuropb_type} with "
                    f"correlation_id: {correlation_id} "
                    f"trace_id: {trace_id} "
                    f": {method.reply_code}, {method.reply_text}"
                )
            )

    def send_message(
        self,
        payload: Dict[str, Any],
        expiry: Optional[int] = None,
        priority: Optional[int] = None,
        encoding: str = "json",
        encrypted: bool = False,
    ) -> None:
        """Send a message to over the RabbitMQ Transport

            TODO: Consider the alternative handling if the channel that's closed.
            - Wait and retry on a new channel?
            - setup a retry queue?
            - should there be a high water mark for the number of retries?
            - should new messages not be consumed until the channel is re-established and retry queue drained?

        :param Dict[str, Any] payload: The message contents
        :param expiry: The message expiry in milliseconds
        :param priority: The message priority
        :param encoding: The encoding of the message
        :param encrypted: True if the message is to be encrypted
        """
        mandatory = False
        exchange = ""
        reply_to = ""

        if payload["tag"] == "event":
            exchange = self._events_exchange
            routing_key = payload["topic"]
        elif payload["tag"] in ("request", "command"):
            mandatory = True
            exchange = self._rpc_exchange
            routing_key = payload["service"]
            reply_to = self._response_queue
        elif payload["tag"] == "response":
            routing_key = payload["reply_to"]
        else:
            raise ValueError(f"Unknown payload type {payload['tag']}")

        if encoding == "json":
            body = encode_payload(
                payload=payload,
                payload_type="json",
            )
        else:
            raise ValueError(f"unsupported encoding {encoding}")

        """ now encrypt the payload if public_key is not None, update RMQ header
        to indicate encrypted payload.
        """
        if (
            encrypted
            and self._encryptor
            and payload["tag"] in ("request", "command", "response")
        ):
            encrypted = "yes"
            to_service = None if payload["tag"] == "response" else payload["service"]
            """ only outgoing response and command messages require the target service name
            """
            wire_body = self._encryptor.encrypt_payload(
                payload=body,
                correlation_id=payload["correlation_id"],
                service_name=to_service,
            )
        else:
            wire_body = body

        properties = dict(
            content_type="application/json",
            correlation_id=payload["correlation_id"],
            reply_to=reply_to,
            headers={
                "nuropb_protocol": NUROPB_PROTOCOL_VERSION,
                "nuropb_version": NUROPB_VERSION,
                "nuropb_type": payload["tag"],
                "trace_id": payload["trace_id"],
                "encrypted": "yes" if encrypted else "",
            },
        )
        if expiry:
            properties["expiration"] = str(expiry)
        if priority:
            properties["priority"] = priority

        if not (self._channel is None or self._channel.is_closed):
            """Check that the channel is in a valid state before sending the response
            # TODO: Consider blocking any new messages until the channel is re-established
            and the message is sent? Especially when it is a response message.
            """
            basic_properties = pika.BasicProperties(**properties)
            self._channel.basic_publish(
                exchange=exchange,
                routing_key=routing_key,
                body=wire_body,
                properties=basic_properties,
                mandatory=mandatory,
            )

        else:
            raise NuropbTransportError(
                description="RMQ channel closed",
                payload=payload,
                exception=None,
            )

    @classmethod
    def acknowledge_service_message(
        cls,
        channel: Channel,
        delivery_tag: int,
        action: Literal["ack", "nack", "reject"],
        redelivered: bool,
    ) -> None:
        """Acknowledgement of a service message

        In NuroPb, acknowledgements of service message requests have three possible outcomes:
        - ack: Successfully processed, acknowledged and removed from the queue
        - nack: A recoverable error occurs, the message is not acknowledged and requeued
        - reject: An unrecoverable error occurs, the message is not acknowledged and dropped

        To prevent accidental use of the redelivered parameter and to ensure system
        predictability on the Call Again feature, messages are only allowed to be redelivered
        once and only once. To this end all messages that have redelivered == True will be
        rejected. if redelivered is overridden with None, it is assumed True.

        :param pika.channel.Channel channel: The channel object
        :param int delivery_tag: The delivery tag
        :param str action: The action to take, one of ack, nack or reject
        :param bool redelivered: True if the message is being requeued / replayed.
        """
        redelivered = True if redelivered is None else redelivered
        if action == "ack":
            channel.basic_ack(delivery_tag=delivery_tag)
        elif action == "nack":
            if redelivered:
                logger.debug(
                    "Redelivered is True, action is overridden from {action} to rejected"
                )
                channel.basic_reject(delivery_tag=delivery_tag, requeue=False)
                raise NuropbCallAgainReject("Redelivered message is rejecting")
            else:
                if verbose:
                    logger.debug(
                        "Redelivered is False, first time nack and requeue is allowed"
                    )
                channel.basic_nack(delivery_tag=delivery_tag, requeue=True)
        elif action == "reject":
            channel.basic_reject(delivery_tag=delivery_tag, requeue=False)
        else:
            raise ValueError(f"Invalid action {action}")

    @classmethod
    def metadata_metrics(cls, metadata: Dict[str, Any]) -> None:
        """Invoked by the transport after a service message has been processed.

        NOTE - METADATA: keep this metadata in sync with across all these methods:
            - on_service_message, on_service_message_complete
            - on_response_message, on_response_message_complete
            - metadata_metrics

        :param metadata: information to drive message processing metrics
        :return: None
        """
        metadata["end_time"] = time.time()
        metadata["duration"] = metadata["end_time"] - metadata["start_time"]
        logger.debug(f"metadata log: {metadata}")

    def on_service_message_complete(
        self,
        channel: Channel,
        basic_deliver: pika.spec.Basic.Deliver,
        properties: pika.spec.BasicProperties,
        private_metadata: Dict[str, Any],
        response_messages: List[TransportRespondPayload],
        acknowledgement: AcknowledgeAction,
    ) -> None:
        """Invoked by the implementation after a service message has been processed.

        This is provided to the implementation as a helper function to complete the message flow.
        The message flow state parameters: channel, delivery_tag, reply_to and private_metadata
        are hidden from the implementation through the use of functools.partial. The interface
        of this function as it appears to the implementation is:
            response_messages:  List[TransportRespondPayload]
            acknowledgement: Literal["ack", "nack", "reject"]

        NOTE: The acknowledgement references the incoming service message that resulted in
              these responses

        :param channel:
        :param pika.spec.Basic.Deliver basic_deliver: basic_deliver method
        :param pika.spec.BasicProperties properties: properties
        :param private_metadata: information to drive message processing metrics
        :param response_messages: List[TransportRespondPayload]
        :param acknowledgement:
        :return:
        """
        delivery_tag = basic_deliver.delivery_tag
        redelivered = basic_deliver.redelivered
        reply_to = properties.reply_to
        trace_id = properties.headers.get("trace_id", "unknown")
        correlation_id = properties.correlation_id
        encrypted = properties.headers.get("encrypted", "") == "yes"

        if redelivered is True and acknowledgement == "nack":
            if verbose:
                logger.debug(
                    "Redelivered is True, action is overridden from nack to rejected"
                )
            acknowledgement = "reject"
            if properties.headers.get("nuropb_type", "") == "request":
                exception = NuropbCallAgainReject(
                    description=(
                        f"Rejecting second call again request for trace_id: {trace_id}"
                        f", correlation_id: {correlation_id}"
                    )
                )
                response_messages[0]["nuropb_payload"][
                    "error"
                ] = error_dict_from_exception(exception=exception)
            else:
                response_messages = []
        elif acknowledgement == "nack":
            response_messages = []

        self.acknowledge_service_message(
            channel, delivery_tag, acknowledgement, redelivered
        )

        for respond_message in response_messages:
            respond_payload = respond_message["nuropb_payload"]
            if respond_payload["tag"] == "response":
                respond_payload["reply_to"] = reply_to
            respond_payload["correlation_id"] = correlation_id
            respond_payload["trace_id"] = trace_id

            self.send_message(
                payload=respond_payload,
                priority=None,
                encoding="json",
                encrypted=encrypted,
            )

        """ NOTE - METADATA: keep this dictionary in sync with across all these methods:
            - on_service_message, on_service_message_complete
            - on_response_message, on_response_message_complete
            - metadata_metrics
        """
        private_metadata["acknowledgement"] = acknowledgement
        private_metadata["message_count"] = len(response_messages)
        private_metadata["flow_complete"] = True
        self.metadata_metrics(metadata=private_metadata)

    def on_service_message(
        self,
        queue_name: str,
        channel: Channel,
        basic_deliver: pika.spec.Basic.Deliver,
        properties: pika.spec.BasicProperties,
        body: bytes,
    ) -> None:
        """Invoked when a message is delivered to the service_queue.

        DESIGN PARAMETERS:
            - Incoming service messages handling (this) require a deliberate acknowledgement.
                * Acknowledgements must be sent on the same channel and with the same delivery tag
            - Incoming response messages (on_response_message) are automatically acknowledged
            - Not all service messages require a response, e.g. events and commands
            - Transport must remain agnostic to the handling of the service message

        OTHER DESIGN CONSIDERATIONS:
            - Connections to the RabbitMQ broker should be authenticated, and encrypted using TLS
            - Payload data not used for NuroPb message handling should be separately encrypted,
              and specially when containing sensitive data e.g. PI or PII .

        FLOW DESIGN:
        1. Python json->dict compatible message is received from the service request (rpc) queue
        2. message is decoded into a python typed dictionary: TransportServicePayload
        3. this message is passed to the message_callback which has no return value, but may raise any
            type of exception.
            - message_callback is provided by the implementation and supplied on transport instantiation
            - message_callback is responsible for handling all exceptions, acknowledgements and
              appropriate responses
            - a helper function is provided that will handle message acknowledgement and any responses.
                the function has all the required state to do so. See note 4 below.
        4. State required to be preserved until the message is acknowledged:
            - channel + delivery_tag
            - reply_to
            - correlation_id
            - trace_id
            - incoming decoded message
        5. This state can only exist in-memory as it's not all "picklable" and bound to the current open channel
        6. If an exception is raised, then the message is immediately rejected and dropped.

        (transport) -> on_service_message
            -> (transport) message_callback(TransportServicePayload, MessageCompleteFunction, Dict[str, Any])
                -> (api) message_complete_callback(List[TransportRespondPayload], AcknowledgeAction)
                    -> (transport) acknowledge_service_message(channel, delivery_tag, action)
                    -> (transport) send_response_messages(reply_to, response_messages)
                    -> (transport) metadata_metrics(metadata)

        # TODO: Needing to think about excessive errors and decide on a strategy for for shutting down
        #       the service instance. Should this take place in the transport layer or the API ?
        # - what happens to the result if the channel is closed?
        # - What happens if the request is resent, can we leverage the existing result
        # - what happens if the request is resent to another service worker?
        # - what happens to the request sender waiting for a response?

        NOTES:
        * Call Again:
        When a message is nack'd and requeued, there is no current way to track how many
        times this may have occurred for the same message. To ensure stability, behaviour
        predictability and to limit the abuse of patterns, the use of NuropbCallAgain is
        limited to once and only once. This is enforced at the transport layer. For RabbitMQ
        if the incoming message is a requeued message, the basic_deliver.redelivered is
        True. Call again for all redelivered messages will be rejected.

        :param str queue_name: The name of the queue that the message was received on
        :param pika.channel.Channel channel: The channel object
        :param pika.spec.Basic.Deliver basic_deliver: basic_deliver method
        :param pika.spec.BasicProperties properties: properties
        :param bytes body: The message body
        """
        nuropb_type = properties.headers.get("nuropb_type", "")
        nuropb_version = properties.headers.get("nuropb_version", "")
        encrypted = properties.headers.get("encrypted", "") == "yes"
        if not verbose:
            logger.debug(f"service message received: '{nuropb_type}'")
        else:
            logger.debug(
                f"""MESSAGE FROM SERVICE QUEUE:
delivery_tag: {basic_deliver.delivery_tag}
from_exchange: {basic_deliver.exchange}
from_queue: {queue_name}
routing_key: {basic_deliver.routing_key}
correlation_id: {properties.correlation_id}
trace_id: {properties.headers.get('trace_id', '')}
content_type: {properties.content_type}
nuropb_type: {nuropb_type}
nuropb_version: {nuropb_version}
encrypted: {encrypted}
"""
            )
        """ Handle for the unknown message type and reply to the originator if possible
        """
        nuropb_version = properties.headers.get("nuropb_version", "")
        correlation_id = properties.correlation_id
        """ NOTE - METADATA: keep this dictionary in sync with across all these methods:
            - on_service_message, on_service_message_complete
            - on_response_message, on_response_message_complete
            - metadata_metrics
        """
        metadata = {
            "start_time": time.time(),
            "service_name": self._service_name,
            "instance_id": self._instance_id,
            "is_leader": self._is_leader,
            "client_only": self._client_only,
            "nuropb_type": nuropb_type,
            "nuropb_version": nuropb_version,
            "correlation_id": correlation_id,
            "trace_id": properties.headers.get("trace_id", "unknown"),
            "encrypted": encrypted,
        }

        message_complete_callback = functools.partial(
            self.on_service_message_complete,
            channel,
            basic_deliver,
            properties,
            metadata,
        )

        """ Decode service message
        """
        try:
            if encrypted and self._encryptor:
                body = self._encryptor.decrypt_payload(
                    payload=body,
                    correlation_id=correlation_id,
                )
            service_message = decode_rmq_body(basic_deliver, properties, body)
        except Exception as error:
            """Exceptions caught here are treated as permanent failures, ack the message and send
            error response is possible
            """
            logger.exception(f"Service message decode error: {error}")
            self.acknowledge_service_message(
                channel, basic_deliver.delivery_tag, "reject", basic_deliver.redelivered
            )

            try:
                (
                    acknowledgement,
                    responses,
                ) = create_transport_response_from_rmq_decode_exception(
                    exception=error, basic_deliver=basic_deliver, properties=properties
                )
                message_complete_callback(responses, acknowledgement)
            except Exception as err:
                logger.debug(
                    f"Failed to send service message decode error response: {err}"
                )

            return

        """ NEXT: handle the service message
        
            Acknowledgement of the message can only take place on this channel and using
            the delivery tag of this message. if the channel is closed before the message is
            acknowledged, then the message will be requeued and redelivered to a new channel
            and potentially a new service worker.

            Any errors caught here are treated as permanent failures, rejected and dropped.        
        """
        try:
            """Assume for now that we can't use key word arguments in self.message_callback"""
            self._message_callback(service_message, message_complete_callback, metadata)

        except Exception as error:
            """Exceptions caught here are treated as permanent failures, ack the message and send error response"""
            logger.exception(f"service message handling error: {error}")
            try:
                (
                    acknowledgement,
                    responses,
                ) = create_transport_response_from_rmq_decode_exception(
                    exception=error, basic_deliver=basic_deliver, properties=properties
                )
                message_complete_callback(responses, acknowledgement)
            except Exception as err:
                logger.debug(f"Failed to send service handling error response: {err}")
            return

    def on_response_message_complete(
        self,
        channel: Channel,
        basic_deliver: pika.spec.Basic.Deliver,
        properties: pika.spec.BasicProperties,
        private_metadata: Dict[str, Any],
        response_messages: List[TransportRespondPayload],
        acknowledgement: AcknowledgeAction,
    ) -> None:
        """Invoked by the implementation after a service message has been processed.

        This is provided to the implementation as a helper function to complete the message flow.
        The message flow state parameters: channel, delivery_tag, reply_to and private_metadata
        are hidden from the implementation through the use of functools.partial. The interface
        of this function as it appears to the implementation is:
            response_messages:  List[TransportRespondPayload]
            acknowledgement: Literal["ack", "nack", "reject"]

        NOTE: The acknowledgement references the incoming service message that resulted in
              these responses

        :param channel:
        :param pika.spec.Basic.Deliver basic_deliver: basic_deliver method
        :param pika.spec.BasicProperties properties: properties
        :param private_metadata: information to drive message processing metrics
        :param response_messages: List[TransportRespondPayload]
        :param acknowledgement:
        :return:
        """
        _ = channel, basic_deliver, properties

        if acknowledgement != "ack":
            logger.warning(
                f"Response messages are auto-acknowledged, ignoring {acknowledgement}"
            )
            acknowledgement = "ack"

        """ Response messages are received from the response queue which is configured for 
        auto-acknowledgement. Code below is for reference only in the event that auto-acknowledgement 
        is disabled or a durable queue is used with explicit acknowledgement.
        
        self.acknowledge_service_message(channel, delivery_tag, acknowledgement, redelivered)
        """

        if len(response_messages) > 0:
            logger.warning(
                "Response messages are not themselves allowed to have response replies, ignoring"
            )

        """ NOTE - METADATA: keep this dictionary in sync with across all these methods:
        - on_service_message, on_service_message_complete
        - on_response_message, on_response_message_complete
        - metadata_metrics
        """
        private_metadata["acknowledgement"] = acknowledgement
        private_metadata["message_count"] = 0
        private_metadata["flow_complete"] = True
        self.metadata_metrics(metadata=private_metadata)

    def on_response_message(
        self,
        _queue_name: str,
        channel: pika.channel.Channel,  # noqa
        basic_deliver: pika.spec.Basic.Deliver,
        properties: pika.spec.BasicProperties,
        body: bytes,
    ) -> None:
        """Invoked when a message is delivered to the response_queue. The channel is passed for
        your convenience. The basic_deliver object that is passed in carries the exchange,
        routing key, delivery tag and a redelivered flag for the message. The properties passed
        in is an instance of BasicProperties with the message properties and the body is the
        message that was sent.

        :param str _queue_name: The name of the queue that the message was received on
        :param pika.channel.Channel channel: The channel object
        :param pika.spec.Basic.Deliver basic_deliver: basic_deliver
        :param pika.spec.BasicProperties properties: properties
        :param bytes body: The message body
        """
        correlation_id = properties.correlation_id
        nuropb_type = properties.headers.get("nuropb_type", "")
        nuropb_version = properties.headers.get("nuropb_version", "")
        encrypted = properties.headers.get("encrypted", "") == "yes"

        if not verbose:
            logger.debug(f"response message received: '{nuropb_type}'")
        else:
            logger.debug(
                (
                    f"MESSAGE FROM RESPONSE QUEUE:\n"
                    f"delivery_tag: {basic_deliver.delivery_tag}\n"
                    f"service_name: {self._service_name}\n"
                    f"instance_id: {self._instance_id}\n"
                    f"exchange: {basic_deliver.exchange}\n"
                    f"routing_key: {basic_deliver.routing_key}\n"
                    f"correlation_id: {correlation_id}\n"
                    f"trace_id: {properties.headers.get('trace_id', '')}\n"
                    f"content_type: {properties.content_type}\n"
                    f"nuropb_type: {nuropb_type}\n"
                    f"nuropb_version: {nuropb_version}\n"
                    f"encrypted: {encrypted}\n"
                )
            )
        try:
            """NOTE - METADATA: keep this dictionary in sync with across all these methods:
            - on_service_message, on_service_message_complete
            - on_response_message, on_response_message_complete
            - metadata_metrics
            """
            metadata = {
                "start_time": time.time(),
                "service_name": self._service_name,
                "instance_id": self._instance_id,
                "is_leader": self._is_leader,
                "client_only": self._client_only,
                "nuropb_type": nuropb_type,
                "nuropb_version": nuropb_version,
                "correlation_id": properties.correlation_id,
                "trace_id": properties.headers.get("trace_id", "unknown"),
                "encrypted": encrypted,
            }
            message_complete_callback = functools.partial(
                self.on_service_message_complete,
                channel,
                basic_deliver.delivery_tag,
                properties.reply_to,
                metadata,
            )
            if encrypted and self._encryptor:
                body = self._encryptor.decrypt_payload(
                    payload=body,
                    correlation_id=correlation_id,
                )
            message = decode_rmq_body(basic_deliver, properties, body)
            self._message_callback(message, message_complete_callback, metadata)
        except Exception as err:
            logger.exception(
                (
                    f"Error processing response message # {basic_deliver.delivery_tag}: {err}\n"
                    f"correlation_id: {properties.correlation_id}\n"
                    f"trace_id: {properties.headers.get('trace_id', '')}\n"
                )
            )

    async def stop_consuming(self) -> None:
        """Tell RabbitMQ that you would like to stop consuming by sending the
        Basic.Cancel RPC command.
        """
        if self._channel:
            logger.info("Sending a Basic.Cancel RPC command to RabbitMQ")
            logger.info("Closing consumers %s", self._consumer_tags)

            all_consumers_closed: Awaitable[bool] = asyncio.Future()

            def _on_cancel_ok(frame: pika.frame.Method) -> None:
                logger.info("Consumer %s closed ok", frame.method.consumer_tag)
                self._consumer_tags.remove(frame.method.consumer_tag)
                if len(self._consumer_tags) == 0:
                    all_consumers_closed.set_result(True)  # type: ignore[attr-defined]

            for consumer_tag in self._consumer_tags:
                if self._channel:
                    self._channel.basic_cancel(consumer_tag, _on_cancel_ok)
            logger.info(
                "Waiting for %ss for consumers to close", CONSUMER_CLOSED_WAIT_TIMEOUT
            )
            try:
                await asyncio.wait_for(
                    all_consumers_closed, timeout=CONSUMER_CLOSED_WAIT_TIMEOUT
                )
            except asyncio.TimeoutError:
                logger.error(
                    "Timed out while waiting for all consumers to gracefully close"
                )

            if len(self._consumer_tags) != 0:
                logger.error(
                    "Timed out while waiting for all consumers to gracefully close"
                )

            self._consuming = False
            logger.info("RabbitMQ acknowledged the cancellation of the consumer")
            self.close_channel()

    def close_channel(self) -> None:
        """Call to close the channel with RabbitMQ cleanly by issuing the Channel.Close RPC command."""
        logger.info("Closing the channel")
        if self._channel is None or self._channel.is_closed:
            logger.info("Channel is already closed")
        else:
            self._channel.close()



import logging
from typing import Any, Dict
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import rsa

from nuropb.contexts.context_manager import NuropbContextManager
from nuropb.contexts.context_manager_decorator import nuropb_context
from nuropb.contexts.describe import publish_to_mesh
from nuropb.interface import NuropbCallAgain, NuropbSuccess

logger = logging.getLogger(__name__)


def get_claims_from_token(bearer_token: str) -> Dict[str, Any] | None:
    """This is a stub for the authorise_func that is used in the tests"""
    _ = bearer_token
    return {
        "sub": "test_user",
        "user_id": "test_user",
        "scope": "openid, profile",
        "roles": "user, admin",
    }


class ServiceExample:
    _service_name: str
    _instance_id: str
    _private_key: rsa.RSAPrivateKey
    _method_call_count: int

    def __init__(self, service_name: str, instance_id: str):
        self._service_name = service_name
        self._instance_id = instance_id
        self._private_key = private_key = rsa.generate_private_key(
            public_exponent=65537, key_size=2048, backend=default_backend()
        )
        self._method_call_count = 0
        self.raise_call_again_error = True

    def test_method(self, **kwargs: Any) -> str:
        _ = kwargs
        self._method_call_count += 1
        return f"response from {self._service_name}.test_method"

    async def test_async_method(self, **kwargs: Any) -> str:
        _ = kwargs
        self._method_call_count += 1
        return f"response from {self._service_name}.test_async_method"

    def test_success_error(self, **kwargs: Any) -> None:
        self._method_call_count += 1
        logger.debug(f"test_success_error: {kwargs}")
        success_result = f"response from {self._service_name}.test_success_error"
        raise NuropbSuccess(
            result=success_result,
        )

    @nuropb_context
    @publish_to_mesh(authorise_func=get_claims_from_token)
    def test_requires_user_claims(self, ctx, **kwargs: Any) -> Any:
        assert isinstance(self, ServiceExample)
        assert isinstance(ctx, NuropbContextManager)
        self._method_call_count += 1
        logger.debug(f"test_requires_user_claims: {kwargs}")
        return ctx.user_claims

    @nuropb_context
    @publish_to_mesh(authorise_func=get_claims_from_token)
    def test_requires_encryption(self, ctx, **kwargs: Any) -> Any:
        assert isinstance(self, ServiceExample)
        assert isinstance(ctx, NuropbContextManager)
        self._method_call_count += 1
        logger.debug(f"test_requires_encryption: {kwargs}")
        return "Result that's to be encrypted in transit"

    def test_call_again_error(self, **kwargs: Any) -> Dict[str, Any]:
        self._method_call_count += 1
        logger.debug(f"test_call_again_error: {kwargs}")
        success_result = f"response from {self._service_name}.test_call_again_error"
        if self.raise_call_again_error:
            """this is preventing the test from getting into an infinite loop"""
            self.raise_call_again_error = False
            raise NuropbCallAgain("Test Call Again")
        result = kwargs.copy()
        result.update(
            {
                "success": success_result,
                "count": self._method_call_count,
            }
        )
        return result

    def test_call_again_loop(self, **kwargs: Any) -> None:
        self._method_call_count += 1
        logger.debug(f"test_call_again_error very time: {kwargs}")
        raise NuropbCallAgain("Test Call Again")



from typing import Dict, Optional
from base64 import b64encode, b64decode

from cryptography.fernet import Fernet
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.asymmetric import padding


def encrypt_payload(payload: str | bytes, key: str | bytes) -> bytes:
    """Encrypt encoded payload with a key
    :param payload: str | bytes
    :param key: str
    :return: bytes
    """
    f = Fernet(key=key, backend=default_backend())
    if isinstance(payload, str):
        payload = payload.encode()
    enc_payload = f.encrypt(payload)
    return b64encode(enc_payload)


def decrypt_payload(encrypted_payload: str | bytes, key: str | bytes) -> bytes:
    """Decrypt encoded payload with a key
    :param encrypted_payload: str | bytes
    :param key: str
    :return: bytes
    """
    f = Fernet(key=key, backend=default_backend())
    payload_to_decrypt = b64decode(encrypted_payload)
    payload = f.decrypt(payload_to_decrypt)
    return payload


def encrypt_key(symmetric_key: bytes, public_key: rsa.RSAPublicKey) -> bytes:
    """Encrypt a symmetric key with an RSA public key
    :param symmetric_key: bytes
    :param public_key: rsa.RSAPublicKey
    :return: bytes, encrypted symmetric key
    """
    encrypted_symmetric_key = public_key.encrypt(
        symmetric_key,
        padding.OAEP(
            mgf=padding.MGF1(algorithm=hashes.SHA256()),
            algorithm=hashes.SHA256(),
            label=None,
        ),
    )
    return b64encode(encrypted_symmetric_key)


def decrypt_key(
    encrypted_symmetric_key: bytes, private_key: rsa.RSAPrivateKey
) -> bytes:
    """Decrypt a symmetric key encrypted with an RSA private key
    :param encrypted_symmetric_key: bytes
    :param private_key: rsa.RSAPrivateKey
    :return: bytes, decrypted symmetric key
    """
    symmetric_key = private_key.decrypt(
        b64decode(encrypted_symmetric_key),
        padding.OAEP(
            mgf=padding.MGF1(algorithm=hashes.SHA256()),
            algorithm=hashes.SHA256(),
            label=None,
        ),
    )
    return symmetric_key


class Encryptor:
    _service_name: str | None
    _private_key: rsa.RSAPrivateKey | None
    _service_public_keys: Dict[str, rsa.RSAPublicKey]
    _correlation_id_symmetric_keys: Dict[str, bytes]

    def __init__(
        self,
        service_name: Optional[str] = None,
        private_key: Optional[rsa.RSAPrivateKey] = None,
    ):
        self._service_name = service_name
        self._private_key = private_key
        self._service_public_keys = {}
        self._correlation_id_symmetric_keys = {}

    @classmethod
    def new_symmetric_key(cls) -> bytes:
        """Generate a new symmetric key
        :return: bytes
        """
        return Fernet.generate_key()

    def add_service_public_key(
        self, service_name: str, public_key: rsa.RSAPublicKey
    ) -> None:
        """Add a public key for a service
        :param service_name: str
        :param public_key: rsa.RSAPublicKey
        """
        self._service_public_keys[service_name] = public_key

    def get_service_public_key(self, service_name: str) -> rsa.RSAPublicKey:
        """Get a public key for a service
        :param service_name: str
        :return: rsa.RSAPublicKey
        """
        return self._service_public_keys.get[service_name]

    def encrypt_payload(
        self,
        payload: bytes,
        correlation_id: str,
        service_name: Optional[str | None] = None,
    ) -> bytes:
        """Encrypt a payload with a symmetric key

        Modes:
        1. Sending a payload to a service (Encrypt)
        4. Sending a response payload from a service (Encrypt)

        :param payload:
        :param correlation_id:
        :param service_name:
        :return: bytes
        """

        """ When service name is provided, it indicates mode 1, else mode 4 
        """
        if service_name is not None and service_name not in self._service_public_keys:
            raise ValueError(
                f"Service public key not found for service: {service_name}"
            )  # pragma: no cover

        if service_name is None:
            # Mode 4, get public key from the private key
            public_key = self._private_key.public_key()
        else:
            # Mode 1, get public key from the destination service's public key
            public_key = self._service_public_keys[service_name]

        if correlation_id not in self._correlation_id_symmetric_keys:
            # Mode 1, generate a new symmetric key and store it for this correlation_id
            key = self.new_symmetric_key()
            self._correlation_id_symmetric_keys[correlation_id] = key
        else:
            # Mode 4, use the original received symmetric key to encrypt key
            # pop it from the dict as it is not required again
            key = self._correlation_id_symmetric_keys.pop(correlation_id)

        encrypted_key = encrypt_key(key, public_key)
        encrypted_payload = encrypt_payload(payload=payload, key=key)

        return b".".join([encrypted_key, encrypted_payload])

    def decrypt_payload(self, payload: bytes, correlation_id: str) -> bytes:
        """Encrypt a payload with a symmetric key

        Modes:
        2. Receiving a response payload from a service (Decrypt)
        3. Receiving a rpc payload (Decrypt)

        :param payload:
        :param correlation_id:
        :return:
        """
        encrypted_key, encrypted_payload = payload.split(b".", 1)
        if correlation_id not in self._correlation_id_symmetric_keys:
            """Mode 3, use public key from the private key to decrypt key"""
            key = decrypt_key(encrypted_key, self._private_key)
            #  remember the key for this correlation_id to encrypt the response
            self._correlation_id_symmetric_keys[correlation_id] = key
        else:
            """Mode 2, use the original symmetric key to decrypt key
            * pop it from the dict as it is not used again
            """
            key = self._correlation_id_symmetric_keys.pop(correlation_id)

        decrypted_payload = decrypt_payload(encrypted_payload, key)

        return decrypted_payload



import asyncio
import logging
from abc import ABC
from typing import (
    Dict,
    Any,
    Callable,
    Optional,
    TypedDict,
    Literal,
    Type,
    Union,
    List,
)

logger = logging.getLogger(__name__)

NUROPB_VERSION = "0.1.3"
NUROPB_PROTOCOL_VERSION = "0.1.1"
NUROPB_PROTOCOL_VERSIONS_SUPPORTED = ("0.1.1",)
NUROPB_MESSAGE_TYPES = (
    "request",
    "response",
    "event",
    "command",
)

NuropbSerializeType = Literal["json"]
NuropbMessageType = Literal["request", "response", "event", "command"]

NuropbLifecycleState = Literal[
    "client-start",
    "client-encode",
    "client-send",
    "service-receive",
    "service-decode",
    "service-handle",
    "service-encode",
    "service-reply",
    "service-ack",
    "client-receive",
    "client-decode",
    "client-handle",
    "client-ack",
    "client-end",
]


class ErrorDescriptionType(TypedDict):
    error: str
    description: Optional[str]
    context: Optional[Dict[str, Any]]


class EventType(TypedDict):
    """For compatibility with better futureproof serialisation support, Any encoded payload type is
    supported.It is encouraged to use a json compatible key/value Type e.g. Dict[str, Any]

    :target: is currently provided here as an aid for the implementation, there are use cases
    where events are targeted to a specified audience or list or targets. In the NuroPb paradigm,
    targets could be individual users or other services. A service would represent the service
    as a whole, NOT any individual instance of that service.

    It is also advised not to use NuroPb for communication between instances of a service.

    Reference the notes on EventPayloadDict
    """

    topic: str
    payload: Any
    target: Optional[List[Any]]


class RequestPayloadDict(TypedDict):
    """Type[RequestPayloadDict]: represents a request that is sent to a service:
    A request has a response, it is acknowledged by the transport layer after the destination
    service has handled the request.

    REMINDER FOR FUTURE: It is very tempting to support the Tuple[...,Any] for the
    param key. There is much broader downstream compatibility in keeping this as a
    dictionary / key-value mapping.
    """

    tag: Literal["request"]
    correlation_id: str
    context: Dict[str, Any]
    trace_id: Optional[str]
    service: str
    method: str
    params: Dict[str, Any]


class CommandPayloadDict(TypedDict):
    """Type[CommandPayloadDict]: represents a command that is sent to a service:
    A command has no response, it is acked immediately by the transport layer. The
    originator of the command has now knowledge of the execution of the command. If
    a response is required, a request should be used.

    A command is useful when used as a type of directed event.

    REMINDER FOR FUTURE: It is very tempting to support the Tuple[...,Any] for the
    param key. There is much broader downstream compatibility in keeping this as a
    dictionary / key-value mapping.
    """

    tag: Literal["command"]
    correlation_id: str
    context: Dict[str, Any]
    trace_id: Optional[str]
    service: str
    method: str
    params: Dict[str, Any]


class EventPayloadDict(TypedDict):
    """Type[EventPayloadDict]: represents an event that is published to a topic:

    :target: is currently provided here as an aid for the implementation, there are use cases
    where events are targeted to a specified audience or list or targets. In the NuroPb paradigm,
    targets could be individual users or other services. A service would represent the service
    as a whole, NOT any individual instance of that service.

    Reference the notes on EventType
    """

    tag: Literal["event"]
    correlation_id: str
    context: Dict[str, Any]
    trace_id: Optional[str]
    topic: str
    event: Any
    target: Optional[List[Any]]


class ResponsePayloadDict(TypedDict):
    """Type[ResponsePayloadDict]: represents a response to a request:"""

    tag: Literal["response"]
    correlation_id: str
    context: Dict[str, Any]
    trace_id: Optional[str]
    result: Any
    error: Optional[Dict[str, Any]]
    warning: Optional[str]
    reply_to: str


PayloadDict = Union[
    ResponsePayloadDict, RequestPayloadDict, CommandPayloadDict, EventPayloadDict
]

ServicePayloadTypes = Union[ResponsePayloadDict, CommandPayloadDict, EventPayloadDict]
ResponsePayloadTypes = Union[ResponsePayloadDict, EventPayloadDict]


class TransportServicePayload(TypedDict):
    """Type[TransportServicePayload]: represents valid service instruction encoded payload.
    Depending on the transport implementation, there wire encoding and serialization may
    be different, and some of the fields may be in the body or header of the message.
    """

    nuropb_protocol: str  # nuropb defined and validated
    correlation_id: str  # nuropb defined
    trace_id: Optional[str]  # implementation defined
    ttl: Optional[int]  # time to live in milliseconds
    nuropb_type: NuropbMessageType
    nuropb_payload: Dict[str, Any]  # ServicePayloadTypes


class TransportRespondPayload(TypedDict):
    """Type[TransportRespondPayload]: represents valid service response message,
    valid nuropb encoded payload types are ResponsePayloadDict, and EventPayloadDict
    """

    nuropb_protocol: str  # nuropb defined and validated
    correlation_id: str  # nuropb defined
    trace_id: Optional[str]  # implementation defined
    ttl: Optional[int]
    nuropb_type: NuropbMessageType
    nuropb_payload: ResponsePayloadTypes


ResultFutureResponsePayload = asyncio.Future[ResponsePayloadDict]
ResultFutureAny = asyncio.Future[Any]


AcknowledgeAction = Literal["ack", "nack", "reject"]
AcknowledgeCallbackFunction = Callable[[AcknowledgeAction], None]
""" AcknowledgeCallbackFunction: represents a callable with the inputs:
    - action: AcknowledgeAction  # one of "ack", "nack", "reject"
"""

MessageCompleteFunction = Callable[
    [List[TransportRespondPayload], AcknowledgeAction], None
]
""" MessageCompleteFunction: represents a callable with the inputs:
    - response_messages: List[TransportRespondPayload]  # the responses to be sent
    - acknowledge_action: AcknowledgeAction  # one of "ack", "nack", "reject"
"""

MessageCallbackFunction = Callable[
    [TransportServicePayload, MessageCompleteFunction, Dict[str, Any]], None
]
""" MessageCallbackFunction: represents a callable with the inputs:
    - message: TransportServicePayload  # the decoded message
    - message_complete: Optional[AcknowledgeCallbackFunction]  # a function that is called to acknowledge the message
    - metadata: Dict[str: Any]  # the context of the message
"""

ConnectionCallbackFunction = Callable[[Type["NuropbInterface"], str, str], None]
""" ConnectionCallbackFunction: represents a callable with the inputs:
    - instance: type of NuropbInterface
    - status: str  # the status of the connection (connected, disconnected)
    - reason: str  # the reason for the connection status change
"""


class NuropbException(Exception):
    """NuropbException: represents a base exception for all exceptions raised by the nuropb API
    although the input parameters are optional, it is recommended that the message is set to a
    meaningful value and the nuropb_message is set to the values that were present when the
    exception was raised.
    """

    description: str
    payload: PayloadDict | TransportServicePayload | TransportRespondPayload | Dict[
        str, Any
    ] | None
    exception: BaseException | None

    def __init__(
        self,
        description: Optional[str] = None,
        payload: Optional[PayloadDict] = None,
        exception: Optional[BaseException] = None,
    ):
        if description is None:
            description = (
                f" {exception}"
                if exception is not None
                else f"{self.__class__.__name__}"
            )
        super().__init__(description)
        self.description = description
        self.payload = payload
        self.exception = exception

    def to_dict(self) -> Dict[str, Any]:
        underlying_exception = str(self.exception) if self.exception else str(self)
        description = self.description if self.description else underlying_exception
        return {
            "error": self.__class__.__name__,
            "description": description,
        }


class NuropbTimeoutError(NuropbException):
    """NuropbTimeoutError: represents an error that occurred when a timeout was reached."""


class NuropbTransportError(NuropbException):
    """NuropbTransportError: represents an error that inside the plumbing."""


class NuropbMessageError(NuropbException):
    """NuropbMessageError: represents an error that occurred during the encoding or decoding of a
    message.
    """


class NuropbHandlingError(NuropbException):
    """NuropbHandlingError: represents an error that occurred during the execution or fulfilment
    of a request or command. An error response is returned to the requester.
    """


class NuropbDeprecatedError(NuropbHandlingError):
    """NuropbDeprecatedError: represents an error that occurred during the execution or fulfilment
    of a request, command or event topic that has been marked deprecated.
    """


class NuropbValidationError(NuropbException):
    """NuropbValidationError: represents an error that occurred during the validation of a
    request or command. An error response is returned to the requester.

    An error response is returned to the requester ONLY for requests and commands.
    Events will be rejected with a NACK with requeue=False.
    """


class NuropbAuthenticationError(NuropbException):
    """NuropbAuthenticationError: when this exception is raised, the transport layer will ACK the
    message and return an authentication error response to the requester.

    This exception occurs whe the identity of the requester can not be validated. for example
    an unknown, invalid or expired user identifier or auth token.

    In most cases, the requester will not be able to recover from this error and will need provide
    valid credentials and retry the request. The approach to this retry outside the scope of the
    nuropb API.
    """


class NuropbAuthorizationError(NuropbException):
    """NuropbAuthorizationError: when this exception is raised, the transport layer will ACK the
    message and return an authorization error response to the requester.

    This exception occurs whe the requester does not have the required privileges to perform the
    requested action of either a request or command.

    In most cases, the requester will not be able to recover from this error and will need provide
    valid credentials and retry the request. The approach to this retry outside the scope of the
    nuropb API.
    """


class NuropbNotDeliveredError(NuropbException):
    """NuropbNotDeliveredError: when this exception is raised, the transport layer will ACK the
    message and return an error response to the requester.
    """


class NuropbCallAgainReject(NuropbException):
    """NuropbCallAgainReject: when this exception is raised, the transport layer will REJECT
    the message

    To prevent accidental use of the redelivered parameter and to ensure system predictability
    on the Call Again feature, messages are only allowed to be redelivered once and only once.
    To this end all messages that have redelivered == True will be rejected.
    """


class NuropbCallAgain(NuropbException):
    """NuropbCallAgain: when this exception is raised, the transport layer will NACK the message
    and schedule it to be redelivered. The delay is determined by the transport layer or message
    broker. A call again will result in a forced repeated call of the original message, with the
    same correlation_id and trace_id.

    The call again "feature" is ignored for event service messages and response messages, as
    these are acked in all cases. The call again feature by implication is only supported for
    request and command messages.

    WARNING: with request messages, if a response has been returned, then this pattern
             SHOULD NOT be used. The requester will receive the same response again, which will be
             ignored as an unpaired response. if the underlying service method has no idempotence
             guarantees, the service could end up in an inconsistent state.
    """


class NuropbSuccess(NuropbException):
    """NuropbSuccessError: when this exception is raised, the transport layer will ACK the message
    and return a success response if service encoded payload is a 'request'. This is useful when the request
    is a command or event and is executed asynchronously.

    There are some use cases where the service may want to return a success response irrespective
    to the handling of the request.

    A useful example is to short circuit processing when an outcome can be predetermined from the
    inputs alone. For end to end request-response consistency, this class must be instantiated with
    ResponsePayloadDict that contains a result consistent with the method and inputs provided.

    Another use case is for the transmission of events raised during the execution of an event,
    command or request. Events will only be sent to the transports layer after the successful
    processing of a service message.
    """

    result: Any
    payload: ResponsePayloadDict | None
    events: List[EventType] = []

    def __init__(
        self,
        result: Any,
        description: Optional[str] = None,
        payload: Optional[ResponsePayloadDict] = None,
        events: Optional[List[EventType]] = None,
    ):
        super().__init__(
            description=description,
            payload=payload,
            exception=None,
        )
        self.result = result
        self.payload = payload
        self.events = [] if events is None else events


class NuropbInterface(ABC):
    """NuropbInterface: represents the interface that must be implemented by a nuropb API implementation"""

    _service_name: str
    _instance_id: str
    _service_instance: object

    @property
    def service_name(self) -> str:
        """service_name: returns the service name"""
        return self._service_name

    @property
    def instance_id(self) -> str:
        """instance_id: returns the instance id"""
        return self._instance_id

    async def connect(self) -> None:
        """connect: waits for the underlying transport to connect, an exception is raised if the connection fails
        to be established
        :return: None
        """
        raise NotImplementedError()  # pragma: no cover

    async def disconnect(self) -> None:
        """disconnect: disconnects from the underlying transport
        :return: None
        """
        raise NotImplementedError()  # pragma: no cover

    @property
    def connected(self) -> bool:
        """connected: returns the connection status of the underlying transport
        :return: bool
        """
        raise NotImplementedError()  # pragma: no cover

    @property
    def is_leader(self) -> bool:
        """is_leader: returns the leader status of the service instance
        :return: bool
        """
        raise NotImplementedError()  # pragma: no cover

    def receive_transport_message(
        self,
        service_message: TransportServicePayload,
        message_complete_callback: MessageCompleteFunction,
        metadata: Dict[str, Any],
    ) -> None:
        """handle_message: does the processing of a NuroPb message received from the transport
        layer.

        All response, request, command and event messages received from the transport layer are
        handled here.

        For failures service messages are handled, other than for events, a response including
        details of the error is returned to the flow originator.

        :param service_message: TransportServicePayload
        :param message_complete_callback: MessageCompleteFunction
        :param metadata: Dict[str, Any] - metric gathering information
        :return: None
        """
        raise NotImplementedError()  # pragma: no cover

    async def request(
        self,
        service: str,
        method: str,
        params: Dict[str, Any],
        context: Dict[str, Any],
        ttl: Optional[int] = None,
        trace_id: Optional[str] = None,
        rpc_response: bool = True,
    ) -> Union[ResponsePayloadDict, Any]:
        """request: sends a request to the target service and waits for a response. It is up to the
        implementation to manage message idempotency and message delivery guarantees.

        :param service: the service name
        :param method: the method name
        :param params: the method arguments, these must be easily serializable to JSON
        :param context: additional information that represent the context in which the request is executed.
            The must be easily serializable to JSON.
        :param ttl: the time to live of the request in milliseconds. After this time and dependent on the
            state and underlying transport, it will not be consumed by the target service and
            should be assumed by the requester to have failed with an undetermined state.
        :param trace_id: an identifier to trace the request over the network (e.g. uuid4 hex string)
        :param rpc_response: if True (default), the actual response of the RPC call is returned and where
            there was an error, that is raised as an exception. Where rpc_response is a ResponsePayloadDict,
             it is returned.

        :return: ResponsePayloadDict
        """
        raise NotImplementedError()  # pragma: no cover

    def command(
        self,
        service: str,
        method: str,
        params: Dict[str, Any],
        context: Dict[str, Any],
        ttl: Optional[int] = None,
        trace_id: Optional[str] = None,
    ) -> None:
        """command: sends a command to the target service. It is up to the implementation to manage message
        idempotency and message delivery guarantees.

        any response from the target service is ignored.

        :param service: the service name
        :param method: the method name
        :param params: the method arguments, these must be easily serializable to JSON
        :param context: additional information that represent the context in which the request is executed.
                        The must be easily serializable to JSON.
        :param ttl: the time to live of the request in milliseconds. After this time and dependent on the
                    underlying transport, it will not be consumed by the target
                    or
                    assumed by the requester to have failed with an undetermined state.
        :param trace_id: an identifier to trace the request over the network (e.g. uuid4 hex string)

        :return: None
        """
        raise NotImplementedError()  # pragma: no cover

    def publish_event(
        self,
        topic: str,
        event: Any,
        context: Dict[str, Any],
        trace_id: Optional[str] = None,
    ) -> None:
        """publish_event: publishes an event to the transport layer. the event sender should not have
        any open transaction that is waiting for a response from this message. It is up to the
        implementation to manage message idempotency and message delivery guarantees.

        :param topic: the topic to publish to
        :param event: the message to publish, must be easily serializable to JSON
        :param context: additional information that represent the context in which the event is executed.
                        The must be easily serializable to JSON.
        :param trace_id: an identifier to trace the request over the network (e.g. uuid4 hex string)
        :return: None
        """
        raise NotImplementedError()  # pragma: no cover



""" This module provides entire nuropb package with json serialisation logic and features
# TODO: Re-check the serialised datetime, date, time and timedelta formats. Look for standards.
"""
from typing import Any, Dict
import json
import datetime
from decimal import Decimal
import dataclasses


def to_json_compatible(obj: Any, recursive: bool = True, max_depth: int = 4) -> Any:
    """Returns a json compatible value for obj, if obj is not a native json type, then
    return a string representation.

    *NOTE 1* This function must be kept in step with the custom json encoder,
    NuropbEncoder, below. Next, read NOTE 2.

    *NOTE 2* This function does not exactly follow the structure of the custom json encoder,
    NuropbEncoder, below. Difference is that the json library implements its own object
    traversal logic. In this function it's required to be done explicitly.

    datetime.datetime: isoformat() + "Z". if there's timezone info, the datetime is
        converted to utc. if there is no timezone info, the datetime is assumed to be utc.

    :param obj: Any
    :param recursive: bool, whether to recursively convert obj to json compatible types
    :param max_depth: int, the maximum depth to recurse
    :return: str, or other json compatible type
    """
    if max_depth < 0:
        return obj  # pragma: no cover

    if isinstance(obj, datetime.datetime):
        if obj.tzinfo is None:
            obj = obj.replace(tzinfo=datetime.timezone.utc)  # assume and set to UTC
        elif obj.tzinfo != datetime.timezone.utc:
            obj = obj.astimezone(datetime.timezone.utc)  # convert to UTC

        json_string = f"{obj.isoformat()}Z"
        return json_string

    if isinstance(obj, (datetime.date, datetime.time)):
        json_string = obj.isoformat()
        return json_string

    if isinstance(obj, datetime.timedelta):
        json_string = str(obj)
        return json_string

    if isinstance(obj, Decimal):
        return "{0:f}".format(obj)

    if dataclasses.is_dataclass(obj):
        v = dataclasses.asdict(obj)
        return to_json_compatible(v, recursive=recursive, max_depth=max_depth - 1)

    if isinstance(obj, dict) and recursive:
        return {
            k: to_json_compatible(v, recursive=recursive, max_depth=max_depth - 1)
            for k, v in obj.items()
        }

    if isinstance(obj, (list, tuple)) and recursive:
        return [
            to_json_compatible(v, recursive=recursive, max_depth=max_depth - 1)
            for v in obj
        ]

    if isinstance(obj, set) and recursive:
        return [
            to_json_compatible(v, recursive=recursive, max_depth=max_depth - 1)
            for v in obj
        ]

    return obj


class NuropbEncoder(json.JSONEncoder):
    """
    *NOTE 1* This class must be kept in step with the function to_json_compatible, above.
    Next, read NOTE 2.

    *NOTE 2* This class does not exactly follow the structure of the function,
    to_json_compatible, above. Difference is that the json library implements its own
    object traversal logic. In the function it's required to be done explicitly.
    """

    def default(self, obj: Any) -> Any:
        if isinstance(obj, datetime.datetime):
            json_string = f"{obj.isoformat()}Z"
            return json_string

        if isinstance(obj, (datetime.date, datetime.time)):
            json_string = f"{obj.isoformat()}Z"
            return json_string

        if isinstance(obj, datetime.timedelta):
            json_string = str(obj)
            return json_string

        if isinstance(obj, Decimal):
            return "{0:f}".format(obj)

        if dataclasses.is_dataclass(obj):
            dataclass_dict = dataclasses.asdict(obj)
            return dataclass_dict

        try:
            return super().default(obj)
        except TypeError:
            return str(obj)


def to_json(obj: Any) -> str:
    """Returns a json string representation of the input object, if not a native json type"""
    return json.dumps(obj, cls=NuropbEncoder)


class JsonSerializor(object):
    """Serializes and deserializes nuropb payloads to and from JSON format."""

    _encryption_keys: Dict[str, Any]
    """ encryption keys related to a given correlation_id
    """

    def __init__(self) -> None:
        """Initializes a new JsonSerializor instance."""
        self._encryption_keys = {}

    def encode(self, payload: Any) -> str:
        """Encodes a nuropb encoded payload to JSON.

        :param payload: Any, The encoded payload to encode.
        :return: str, The JSON-encoded encoded payload.
        """
        _ = self
        json_payload = to_json(payload)
        return json_payload

    def decode(self, json_payload: str) -> Any:
        """Decodes a JSON-encoded nuropb encoded payload.

        :param json_payload: str, The JSON-encoded encoded payload to decode.
        :return: Any, The decoded encoded payload.
        """
        _ = self
        payload = json.loads(json_payload)
        return payload



from cryptography.hazmat.primitives.asymmetric import rsa

from nuropb.encodings.json_serialisation import JsonSerializor
from nuropb.interface import PayloadDict


SerializorTypes = JsonSerializor


def get_serializor(payload_type: str = "json") -> SerializorTypes:
    """Returns a serializor object for the specified encoded payload type
    :param payload_type: "json"
    :return: a serializor object
    """
    if payload_type != "json":
        raise ValueError(f"payload_type {payload_type} is not supported")

    return JsonSerializor()


def encode_payload(
    payload: PayloadDict,
    payload_type: str = "json",
    public_key: rsa.RSAPublicKey = None,
) -> bytes:
    """
    :param public_key:
    :param payload:
    :param payload_type: "json"
    :param public_key: rsa.PublicKey
    :return: a json bytes string imputed encoded payload
    """
    if payload_type != "json":
        raise ValueError(f"payload_type {payload_type} is not supported")

    return (
        get_serializor(
            payload_type=payload_type,
        )
        .encode(payload)  # to json
        .encode()  # to bytes
    )


def decode_payload(
    encoded_payload: bytes,
    payload_type: str = "json",
) -> PayloadDict:
    """
    :param encoded_payload:
    :param payload_type: "json"
    :return: PayloadDict
    """
    if payload_type != "json":
        raise ValueError(f"payload_type {payload_type} is not supported")

    payload = get_serializor(
        payload_type=payload_type,
    ).decode(encoded_payload.decode())
    if not isinstance(payload, dict):
        raise ValueError(
            f"Decoded payload is not a dictionary: {type(payload).__name__}"
        )
    return payload



import logging
from types import TracebackType
from typing import Any, Dict, List, Optional, Type

logger = logging.getLogger(__name__)

_test_token_cache: Dict[str, Any] = {}
_test_user_id_cache: Dict[str, Any] = {}


class NuropbContextManager:
    """This class is a context manager that can be used to manage the context transaction relating
    to an incoming nuropb service message. when a class instance method is decorated with the
    nuropb_context decorator, the context manager is instantiated and injected into the method
    as a ctx parameter.

    The nuropb context manager is both a sync and async context manager, it also provides access
    to a decorated method to the nuropb service mesh api. Events can be added to the context
    manager and will be sent to the service mesh when the context manager exits. If an exception
    is raised while the context manager is running, the exception is recorded and the context
    transaction is considered to have failed. If any events were added to the context manager,
    they are discarded.
    """

    _suppress_exceptions: bool
    _nuropb_payload: Dict[str, Any] | None
    _context: Dict[str, Any]
    _user_claims: Dict[str, Any] | None
    _events: List[Dict[str, Any]]
    _exc_type: Type[BaseException] | None
    _exec_value: BaseException | None
    _exc_tb: TracebackType | None
    _started: bool
    _done: bool

    def __init__(
        self, context: Dict[str, Any], suppress_exceptions: Optional[bool] = True
    ):
        if context is None:
            raise TypeError("context cannot be None")
        self._suppress_exceptions = (
            True if suppress_exceptions is None else suppress_exceptions
        )
        self._nuropb_payload = None
        self._context = context
        self._user_claims = None
        self._events = []

        self._exc_type = None
        self._exec_value = None
        self._exc_tb = None
        self._started = False
        self._done = False

    @property
    def context(self) -> Dict[str, Any]:
        return self._context

    @property
    def user_claims(self) -> Dict[str, Any] | None:
        return self._user_claims

    @user_claims.setter
    def user_claims(self, claims: Dict[str, Any] | None) -> None:
        if self._user_claims is not None:
            raise ValueError("user_claims can only be set once")
        self._user_claims = claims

    @property
    def events(self) -> List[Dict[str, Any]]:
        return self._events

    @property
    def error(self) -> Dict[str, Any] | None:
        if self._exc_type is None:
            return None
        return {
            "error": self._exc_type.__name__,
            "description": str(self._exec_value),
        }

    def add_event(self, event: Dict[str, Any]) -> None:
        """Add an event to the context manager. The event will be sent to the service mesh when
        the context manager exits successfully.

        Event format:
        {
            "topic": "test_topic",
            "event": "test_event_payload",
            "context": {}
        }

        :param event:
        :return:
        """
        self._events.append(event)

    def _handle_context_exit(
        self,
        exc_type: Type[BaseException] | None,
        exc_value: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> bool:
        """This method is for customising the behaviour when a context manager exits. It has the
        same signature as __exit__ or __aexit__.

        If an exception was raised while the context manager was running, the exception information
        is recorded and content transaction is considered to have failed. If any events were added
        to the context manager, they are discarded.
        """
        if self._done:
            raise RuntimeError("Context manager has already exited")
        self._done = True
        self._exc_type = exc_type
        self._exec_value = exc_value
        self._exc_tb = exc_tb

        if exc_type is not None:
            self._events = []

        return self._suppress_exceptions

    """ 
        **** Context Manager sync and async methods ****
    """

    def __enter__(self) -> Any:
        """This method is called when entering a context manager with a with statement"""
        if self._done:
            raise RuntimeError("Context manager has already exited")
        if self._started:
            raise RuntimeError("Context manager has already entered")
        self._started = True

        return self

    async def __aenter__(self) -> Any:
        return self.__enter__()

    def __exit__(
        self,
        exc_type: Type[BaseException] | None,
        exc_value: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> bool | None:
        return self._handle_context_exit(exc_type, exc_value, exc_tb)

    async def __aexit__(
        self,
        exc_type: Type[BaseException] | None,
        exc_value: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> bool | None:
        return self.__exit__(exc_type, exc_value, exc_tb)



""" This module provides the runtime configuration for the RabbitMQ transport.
- NuroPb Service leader election
- RMQ Exchange and Queue configuration

A service Leader is elected using etcd. The leader is responsible for creating the RMQ Exchange and Queues,
and binding the Queues to the Exchange. Due to the additional administrative responsibilities, the leader's
message prefetch size should be smaller relative to the other service instances.

TODO: The service leader is also responsible for monitoring and managing overall service health.
- if the service queue is not draining fast enough, the leader will signal for additional instances to be started.
- if the service queue is draining too fast, the leader will signal for instances to be stopped.
- Connections to RMQ that are initiating service messages with high error rates will be terminated.
- The leader will also monitor the dead letter queue and take action if the dead letter queue is growing too fast.

NuroPb services instances wait for the elected leader to signal that the RMQ Exchange and Queues are ready before
connecting and starting to consume messages from the RMQ Queues.
"""
import asyncio
import logging
from asyncio import AbstractEventLoop
from typing import Dict, Any, Awaitable, Literal, Optional

from etcd3 import Client as EtcdClient
from etcd3.stateful.lease import Lease
from etcd3.stateful.watch import Watcher, Event
from etcd3.stateful.transaction import Txn
from requests.exceptions import HTTPError
from pika.exceptions import AMQPConnectionError

from nuropb.rmq_api import RMQAPI
from nuropb.rmq_lib import configure_nuropb_rmq, create_virtual_host
from nuropb.rmq_transport import RMQTransport

logger = logging.getLogger(__name__)

LEADER_KEY = "/leader"  # this key is prefixed with /nuropb/{service-name}
LEASE_TTL = 15  # seconds


class ServiceRunner:
    """ServiceRunner represents the state of the service configuration."""

    service_name: str
    """ service_name: the name of the service. """
    leader_id: str
    """ service_leader_id: the identifier of the service leader. """
    configured: bool
    """ rmq_configured: indicates if the RMQ Exchange and Queues have been configured. """
    ready: bool
    """ rmq_ready: indicates if the RMQ Exchange and Queues are ready to receive messages. """
    consume: bool
    """ rmq_consume: indicates if all service instance should consume messages or not from the RMQ Queues. """
    hw_mark: int
    """ trigger for when to start or stop service instances. """

    _etc_client: EtcdClient
    """ _etc_client: the etcd3 client used to communicate with the etcd3 server. """


ContainerRunningState = Literal[
    "startup",
    "startup-leader-election",
    "startup-leader-decided",
    "taking-leadership",
    "running-leader",
    "running-follower",
    "running-standalone",
    "stopping",
    "shutdown",
]


class ServiceContainer(ServiceRunner):
    _instance: RMQAPI
    _rmq_api_url: str
    _service_name: str
    _transport: RMQTransport
    _running_state: ContainerRunningState
    _rmq_config_ok: bool

    _shutdown: bool
    _is_leader: bool
    _leader_reference: str
    _etcd_config: Dict[str, Any]
    _etcd_client: EtcdClient
    _etcd_lease: Lease
    _etcd_watcher: Watcher
    _etcd_prefix: str

    _container_running_future: Awaitable[bool] | None
    _container_shutdown_future: Awaitable[bool] | None

    def __init__(
        self,
        rmq_api_url: str,
        instance: RMQAPI,
        etcd_config: Optional[Dict[str, Any]] = None,
    ):
        """__init__: initializes the service container.
        :param instance: the service instance.
        :param etcd_config: the etcd3 configuration.
        """
        self._running_state = "startup"
        self._container_running_future = None
        self._container_shutdown_future = None
        self._rmq_config_ok = False
        self._shutdown = False
        self._rmq_api_url = rmq_api_url
        self._instance = instance
        self._service_name = instance.service_name
        self._instance_id = instance.instance_id
        self._transport = instance.transport
        self._is_leader = False
        self._leader_reference = ""
        self._etcd_prefix = f"/nuropb/{self._service_name}"
        self._etcd_config = etcd_config if etcd_config is not None else {}

        logger.info(
            "Starting the service container for {}:{}".format(
                self._service_name, self._instance_id
            )
        )

        if not self._etcd_config:
            logger.warning("etcd features are disabled")
            self.running_state = "running-standalone"
        else:
            """asyncio NOTE: the etcd3 client is initialized as an asyncio task and will run
            once __init__ has completed and there us a running asyncio event loop.
            """
            task = asyncio.create_task(self.init_etcd(on_startup=True))
            task.add_done_callback(lambda _: None)

    @property
    def running_state(self) -> ContainerRunningState:
        """running_state: the current running state of the service container."""
        return self._running_state

    @running_state.setter
    def running_state(self, value: ContainerRunningState) -> None:
        """running_state: updating the current running state of the service container."""
        if value != self._running_state:
            logger.info(f"running_state change: {self._running_state} -> {value}")
            self._running_state = value

    async def init_etcd(self, on_startup: bool = True) -> bool:
        """init_etcd: initialises the etcd client and connection.
        Not a fan of the threading started here, when defining a watcher and calling
        runDaemon(). It would be way nicer to have a fully asynchronous etcd3 client.
        There is also threading used for the lease keepalive.

        NOTES:
         - this function will run until successful etcd connection is established, or
              until self._shutdown is set to True.

        :param on_startup: True, if the etcd3 client is being initialised at startup.
        :return: None
        """
        if on_startup:
            logger.info("Initiating the etcd connection")
            self.running_state = "startup-leader-election"
        else:
            logger.info("Re-initiating the etcd connection")

            """ Potential the exceptions raised while stopping the watcher or expiring the lease
            are deliberately ignored. 
            """
            try:
                if self._etcd_watcher and self._etcd_watcher.watching:
                    self._etcd_watcher.stop()
            except Exception:  # noqa
                pass
            try:
                if self._etcd_lease and self._etcd_lease.keeping:
                    self._etcd_lease.cancel_keepalive()
                    self._etcd_lease.revoke()
            except Exception:  # noqa
                pass

        while True:
            try:
                self._etcd_client = EtcdClient(**self._etcd_config)
                self._etcd_lease = self._etcd_client.Lease(LEASE_TTL)
                self._etcd_lease.grant()
                self._etcd_lease.keepalive()
                logger.info(
                    f"etcd lease {self._etcd_lease.ID} created, with ttl of {self._etcd_lease.ttl()}"
                )
                self._etcd_watcher = self._etcd_client.Watcher(
                    all=True, progress_notify=True, prev_kv=True
                )
                self._etcd_watcher.onEvent(
                    f"{self._etcd_prefix}/.*", self.etcd_event_handler
                )
                self._etcd_watcher.runDaemon()
                self.nominate_as_leader()
                return True
            except Exception as e:
                logger.error(f"Error while initiating etcd client: {e}")
                logger.info("Retrying the etcd client connection in 5 seconds")
                if self._shutdown:
                    break
                else:
                    await asyncio.sleep(5)

        return False

    def nominate_as_leader(self) -> None:
        """nominate_as_leader: nominates the service instance as the leader. if first to secure a lease
        on the leader key, then the instance is the leader. Otherwise, the instance is a follower.

        Handling the following scenarios:
            On Startup: [No transition to leader or from leader required]
            * Can secure the leader key lease
                - startup: the first instance to start, no existing entry becomes the leader.
            * Cannot secure the leader key lease
                - startup: there is an existing leader entry, all services must wait until key's lease expires
                    and can then compete to become the leader.

            Post Startup: [Possible transition to leader or from leader is required]
            * No change to self._is_leader and continue as before. - No transition required
            * Can secure the leader key lease
                - running: the leader entry is reset to None, followers then compete to become leader.

        ETCD NOTES: it appears that a lease can only be associated with a key it that key was created with
        the lease. If the key is created without a lease, then the lease cannot be associated with the key.

        :return: None
        """
        try:
            if self._instance.client_only:
                logger.info(
                    "Instance configured as a client only, no leader election required"
                )
                self._is_leader = False
                self.running_state = "startup-leader-decided"
                return

            key = f"{self._etcd_prefix}/leader"
            logger.info(f"leader key: {key}")

            txn = Txn(self._etcd_client)
            txn.compare(txn.key(key).version == 0)
            txn.success(txn.put(key, self._instance_id, self._etcd_lease.ID))
            txn.failure(txn.range(key))
            response = txn.commit()
            if response.succeeded:
                if self._leader_reference == self._instance_id:
                    # remains the leader nothing to do
                    return
                logger.debug(f"nomination of self:{self._instance_id} succeeded")
                self._leader_reference = self._instance_id
            else:
                """Unable to secure the leader key lease, check for the assigned leader."""
                if len(response.responses) >= 1:
                    value = response.responses[0].response_range.kvs[0].value.decode()
                    if value == self._instance_id:
                        raise ValueError(
                            "Unable to secure the leader key lease, but the leader is self"
                        )
                    self._leader_reference = value

            is_leader = self._leader_reference == self._instance_id

            """ ON STARTUP LEADER ELECTION
            """
            if self.running_state == "startup-leader-election":
                if is_leader:
                    logger.critical("Elected as leader")
                else:
                    logger.info(
                        f"Instance configured as a follower. Leader instance_id: {self._leader_reference}"
                    )
                self._is_leader = is_leader
                self.running_state = "startup-leader-decided"
                return

            """ POST STARTUP LEADER ELECTION
            """
            if is_leader == self._is_leader:
                """no change to self._is_leader and continue as before. In this case, an update to the
                leadership has no impact on this service instance.
                """
                return

            if is_leader and self._is_leader is False:
                """Elected as leader, handle the transition to leader"""
                logger.critical(
                    "Post-startup election as leader, handle the transition"
                )
                self.running_state = "taking-leadership"
                self._is_leader = is_leader
                # TODO: handle the transition from follower to leader
                return
            else:
                """Transition from leader to follower"""
                logger.critical("Post startup transition from leader to follower")
                self._is_leader = is_leader
                # TODO: handle the transition from leader to follower

        except Exception as e:
            """Error Likely to be due to etcd server not being available. In this case do nothing
            to the leadership state and wait for the next etcd event or reconnection attempt.
            """
            logger.error(f"Error during leader nomination: {e}")
            logger.exception(e)
            logger.info("called init_etcd() to re-initiate the etcd connection")
            task = asyncio.create_task(self.init_etcd(on_startup=False))
            task.add_done_callback(lambda _: None)

    def update_etcd_service_property(self, key: str, value: Any) -> bool:
        """update_etcd_service_property: updates the etcd3 service property.

        The use of this method is restricted to the service leader

        :param key:
        :param value:
        :return:
        """
        key = f"{self._etcd_prefix}/{key}"
        try:
            if self._is_leader:
                self._etcd_client.put(key, value)
                return True
        except Exception as e:
            logger.error(f"error updating etcd service property {key}: {e}")

        return False

    def check_and_configure_rmq(self) -> None:
        """check_and_configure_rmq: checks that the RMQ Exchange and Queues are correctly
        configured. If not, then the NuroPb RMQ configuration for self.service_name is applied.
        :return: None
        """
        amqp_url: str = self._transport.amqp_url
        create_virtual_host(self._rmq_api_url, amqp_url)
        transport_settings = self._transport.rmq_configuration
        configure_nuropb_rmq(
            service_name=self._service_name,
            rmq_url=amqp_url,
            events_exchange=transport_settings["events_exchange"],
            rpc_exchange=transport_settings["rpc_exchange"],
            dl_exchange=transport_settings["dl_exchange"],
            dl_queue=transport_settings["dl_queue"],
            service_queue=transport_settings["service_queue"],
            rpc_bindings=transport_settings["rpc_bindings"],
            event_bindings=transport_settings["event_bindings"],
        )
        self._rmq_config_ok = True

    def etcd_event_handler(self, event: Event) -> None:
        """etc_event_handler: handles the etcd3 events."""
        logger.critical(f"WATCHER: key change {event.key}: {event.value}")
        event_key = event.key.decode()
        # logger.info(f"{event_key} - {self._etcd_prefix}/leader")
        if event_key == f"{self._etcd_prefix}/leader":
            """Only nominate for a new leader when the leader key is reset to None"""
            new_reference = event.value.decode() if event.value else None
            # logger.info(f"new_reference: {new_reference}")
            if new_reference is None:
                # there is no leader currently so then compete for leadership
                self.nominate_as_leader()
        elif event.value and event_key == f"{self._etcd_prefix}/rmq-config-ok":
            self._rmq_config_ok = event.value.decode() == "True"

    async def startup_steps(self) -> None:
        """startup_steps: the startup steps for the service container.
        - wait for the leader to be elected.
        - check and configure the RMQ Exchange and Queues.
        - connect the service instance to RMQ.
        - start the etcd lease refresh loop.
        :return: None
        """
        while True:
            if self.running_state in ("startup-leader-decided", "running-standalone"):
                break
            else:
                await asyncio.sleep(1)

        if self.running_state == "running-standalone":
            self.check_and_configure_rmq()
            self._rmq_config_ok = True

        elif self._is_leader:
            self.check_and_configure_rmq()
            self.update_etcd_service_property("rmq-config-ok", self._rmq_config_ok)

        if self.running_state != "running-standalone":
            """For all service instances including the leader, wait for the leader to confirm that the
            RMQ configuration to be OK before connecting to RMQ. Although the RMQ configuration queue
            is idempotent, it is better consistency across the service instances to wait.
            """
            if self._instance.client_only is False and self._rmq_config_ok is False:
                response = self._etcd_client.range(f"{self._etcd_prefix}/rmq-config-ok")
                if len(response.kvs) >= 1 and response.kvs[0].value:
                    self._rmq_config_ok = response.kvs[0].value.decode() == "True"
                while self._rmq_config_ok is False:
                    logger.debug("Waiting for the RMQ configuration to be OK")
                    await asyncio.sleep(5)  # wait 5 more seconds
            self.running_state = (
                "running-leader" if self._is_leader else "running-follower"
            )

        await self._instance.connect()

    async def start(self) -> bool:
        """start: starts the container service instance.
        - primary entry point to start the service container.
        :return: None
        """
        started = False
        try:
            await self.startup_steps()
            logger.info("Container startup complete")
            started = True
        except AMQPConnectionError as err:
            logger.error(f"Startup error connecting to RabbitMQ: {err}")
        except HTTPError as err:
            logger.error(f"HTTP error calling RabbitMQ API: {err}")
        except ValueError as err:
            logger.error(f"Startup error, likely due to miss configuration: {err}")
        except (asyncio.CancelledError, Exception) as err:
            if isinstance(err, asyncio.CancelledError):
                logger.info(f"container running future cancelled: {err}")
            else:
                logger.exception(f"Container running future runtime exception: {err}")
        finally:
            return started

    async def stop(self) -> None:
        """stop: stops the container service instance.
        - primary entry point to stop the service container.
        :return: None
        """
        self.running_state = "stopping"
        self._container_shutdown_future = asyncio.Future()
        self._shutdown = True
        self._etcd_watcher.stop()
        try:
            await self._container_shutdown_future
            self.running_state = "shutdown"
            logger.info("Container shutdown complete")
        except (asyncio.CancelledError, Exception) as err:
            if isinstance(err, asyncio.CancelledError):
                logger.info(f"container shutdown future cancelled: {err}")
            else:
                logger.exception(f"Container shutdown future runtime exception: {err}")
        finally:
            ioloop: AbstractEventLoop = asyncio.get_running_loop()
            if ioloop.is_running():
                ioloop.stop()



import inspect
from typing import Optional, Any, Callable
from functools import wraps

from nuropb.contexts.context_manager import NuropbContextManager


def method_requires_nuropb_context(method: Callable[..., Any]) -> bool:
    """This function checks if a method has been decorated with @nuropb_context
    :param method: Callable
    :return: bool
    """
    return getattr(method, "__nuropb_context__", False)


def nuropb_context(
    original_method: Optional[Callable[..., Any]] = None,
    *,
    context_parameter: str = "ctx",
    suppress_exceptions: bool = False,
) -> Any:
    """This decorator function injects a NuropbContext instance into a method that has ctx:NuropbContext
    as an argument. The ctx parameter of the decorated method is hidden from the method's signature visible
    on the service mesh.

    The name of the ctx parameter can be changed by passing a string to the context_parameter argument.

    The caller of class_instance.method(ctx=ctx) can either pass a NuropbContext instance or a dict. If a dict
    is passed, a NuropbContext instance will be created from the dict.

    *NOTE* This decorator is only for with class methods, using with functions will have unexpected
    results and is likely to raise a TypeException

    As illustrated by the example below, @nuropb_context must always be on top of @publish_to_mesh when
    both decorators are used.

        @nuropb_context
        @publish_to_mesh(context_token_key="Authorization", authorise_func=authorise_token)
        def hello_requires_auth(...)


    :param original_method: the method to be decorated
    :param context_parameter: str
    :param suppress_exceptions: bool
    :return: a decorated method
    """
    context_parameter = context_parameter or "ctx"
    suppress_exceptions = False if suppress_exceptions is None else suppress_exceptions

    def decorator(method: Callable[..., Any]) -> Callable[..., Any]:
        if context_parameter not in inspect.signature(method).parameters:
            raise TypeError(
                f"method {method.__name__} does not have {context_parameter} as an argument"
            )

        @wraps(method)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            """validate calling arguments"""
            if len(args) < 2 or not isinstance(args[1], (NuropbContextManager, dict)):
                raise TypeError(
                    f"The @nuropb_context, expects a {context_parameter}: NuropbContextManager "
                    f"or dict, as the first argument in calling instance.method"
                )
            method_args = list(args)
            ctx = method_args[1]
            if not isinstance(ctx, NuropbContextManager):
                """Replace dictionary context with NuropbContextManager instance"""
                ctx = NuropbContextManager(
                    context=method_args[1],
                    suppress_exceptions=suppress_exceptions,
                )
                method_args[1] = ctx

            authorise_func = getattr(wrapper, "__nuropb_authorise_func__", None)
            if authorise_func is not None:
                context_token_key = getattr(wrapper, "__nuropb_context_token_key__")
                ctx.user_claims = authorise_func(ctx.context[context_token_key])

            # kwargs[context_parameter] = ctx
            # return method(*args[1:], **kwargs)
            return method(*method_args, **kwargs)

        setattr(wrapper, "__nuropb_context__", True)
        setattr(wrapper, "__nuropb_context_arg__", context_parameter)
        return wrapper

    if original_method:
        return decorator(original_method)
    else:
        return decorator



import logging
import inspect
from typing import Optional, Any, get_origin, get_args, Callable, Dict, Tuple
from functools import wraps
from cryptography.hazmat.primitives import serialization


logger = logging.getLogger(__name__)

AuthoriseFunc = Callable[[str], Dict[str, Any]]


def method_visible_on_mesh(method: Callable[..., Any]) -> bool:
    """This function checks if a method has been decorated with @publish_to_mesh
    :param method: callable
    :return: bool
    """
    return getattr(method, "__nuropb_mesh_visible__", False)


def publish_to_mesh(
    original_method: Optional[Callable[..., Any]] = None,
    *,
    hide_method: Optional[bool] = False,
    authorise_func: Optional[AuthoriseFunc] = None,
    context_token_key: Optional[str] = "Authorization",
    requires_encryption: Optional[bool] = False,
    description: Optional[str] = None,
) -> Any:
    """Decorator to expose class methods to the service mesh

    When a service instance is connected to a service mesh via the service mesh client, all
    the standard public methods of the service instance is available to the service mesh. Methods
    that start with underscore will always remain hidden. methods that are explicitly marked as
    hidden by publish_to_mesh will also not be published.

    When and authorise_func is specified, this function will be called with the contents of
    context[context_token_key]. if the token validation is unsuccessful, then a failed authorisation
    exception is raised. If successful then ctx.user_claims is populated with claims attached to the
    token.

    When requires_encryption is True, the service mesh will encrypt the payload of the service message
    request and response. It is the responsibility of the process making the request to ensure that
    it has the target service's public key.

    As illustrated by the example below, @nuropb_context must always be on top of @publish_to_mesh when
    both decorators are used.

        @nuropb_context
        @publish_to_mesh(context_token_key="Authorization", authorise_func=authorise_token)
        def hello_requires_auth(...)


    :param original_method:
    :param hide_method:
    :param authorise_func:
    :param context_token_key:
    :param requires_encryption:
    :param description: str, if present then override the methods doc string
    :return:
    """
    hide_method = False if hide_method is None else hide_method
    context_token_key = (
        "Authorization" if context_token_key is None else context_token_key
    )
    if authorise_func is not None and not callable(authorise_func):
        raise TypeError("Authorise function must be callable")
    requires_encryption = False if requires_encryption is None else requires_encryption

    def decorator(method: Callable[..., Any]) -> Callable[..., Any]:
        @wraps(method)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            return method(*args, **kwargs)

        setattr(wrapper, "__nuropb_mesh_hidden__", hide_method)
        setattr(wrapper, "__nuropb_context_token_key__", context_token_key)
        setattr(wrapper, "__nuropb_authorise_func__", authorise_func)
        setattr(wrapper, "__nuropb_requires_encryption__", requires_encryption)
        if description:
            setattr(wrapper, "__nuropb_description__", description)
        return wrapper

    if original_method:
        return decorator(original_method)
    else:
        return decorator


def describe_service(class_instance: object) -> Dict[str, Any] | None:
    """Returns a description of the class methods that will be exposed to the service mesh"""
    if class_instance is None:
        logger.warning("No service class base has been input")
        return None
    else:
        service_name = getattr(class_instance, "_service_name", None)
        service_description = getattr(class_instance, "__doc__", None)
        service_description = (
            service_description.strip() if service_description else service_description
        )
        service_version = getattr(class_instance, "_version", None)
        methods = []
        service_has_encrypted_methods = False

        for name, method in inspect.getmembers(class_instance):
            """all private methods are excluded, regardless if one has been decorated with @publish_to_mesh"""
            if name[0] == "_" or not callable(method):
                continue

            if getattr(method, "__nuropb_mesh_hidden__", False):
                continue

            requires_encryption = getattr(
                method, "__nuropb_requires_encryption__", False
            )
            if requires_encryption:
                service_has_encrypted_methods = True

            ctx_arg_name = getattr(method, "__nuropb_context_arg__", "ctx")
            method_signature = inspect.signature(method)
            required = []

            def map_annotation(arg_props: Any) -> str:
                annotation = arg_props.annotation
                label = ""
                if annotation != inspect._empty:
                    origin = get_origin(annotation)
                    if origin is not None:
                        args = [
                            a
                            for a in get_args(annotation)
                            if a.__name__ not in ("NoneType",)
                        ]
                        if len(args) == 1:
                            label = args[0].__name__

                    else:
                        label = annotation.__name__
                return label

            def map_default(arg_props: Any) -> Any:
                default = arg_props.default
                if default == inspect._empty:
                    required.append(arg_props.name)
                    return ""
                else:
                    return default

            def map_argument(arg_props: Any) -> Tuple[str, Dict[str, Any]]:
                return (
                    arg_props.name,
                    {
                        "type": map_annotation(arg_props),
                        "description": "",
                        "default": map_default(arg_props),
                    },
                )

            properties = [
                map_argument(p)
                for n, p in method_signature.parameters.items()
                if n not in ("self", "cls", ctx_arg_name)
            ]

            method_spec = {
                "description": getattr(
                    method, "__nuropb_description__", inspect.getdoc(method)
                ),
                "requires_encryption": requires_encryption,
                "parameters": {
                    "type": "object",
                    "properties": properties,
                    "required": required,
                },
            }
            methods.append((name, method_spec))

        service_info = {
            "service_name": service_name,
            "service_version": service_version,
            "description": service_description,
            "encrypted_methods": service_has_encrypted_methods,
            "methods": dict(methods),
        }

        if service_has_encrypted_methods:
            private_key = service_name = getattr(class_instance, "_private_key", None)
            if private_key is None:
                raise ValueError(
                    f"Service {service_name} has encrypted methods but no private key has been set"
                )

            service_info["public_key"] = (
                private_key.public_key()
                .public_bytes(
                    encoding=serialization.Encoding.PEM,
                    format=serialization.PublicFormat.SubjectPublicKeyInfo,
                )
                .decode("ascii")
            )

        return service_info



import logging
from typing import Dict, Optional, Any, Union, cast
from uuid import uuid4
from asyncio import Future
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.backends import default_backend

from nuropb.encodings.encryption import Encryptor
from nuropb.interface import (
    NuropbInterface,
    NuropbMessageError,
    ResponsePayloadDict,
    RequestPayloadDict,
    EventPayloadDict,
    NuropbHandlingError,
    ResultFutureResponsePayload,
    TransportServicePayload,
    MessageCompleteFunction,
    CommandPayloadDict,
    NuropbException,
)
from nuropb.rmq_transport import RMQTransport
from nuropb.service_handlers import execute_request, handle_execution_result

logger = logging.getLogger(__name__)
verbose = False


class RMQAPI(NuropbInterface):
    """RMQAPI: A NuropbInterface implementation that uses RabbitMQ as the underlying transport."""

    _mesh_name: str
    _connection_name: str
    _response_futures: Dict[str, ResultFutureResponsePayload]
    _transport: RMQTransport
    _rpc_exchange: str
    _events_exchange: str
    _service_instance: object | None
    _default_ttl: int
    _client_only: bool
    _encryptor: Encryptor
    _service_discovery: Dict[str, Any]
    _service_public_keys: Dict[str, Any]

    def __init__(
        self,
        amqp_url: str,
        service_name: str | None = None,
        instance_id: str | None = None,
        service_instance: object | None = None,
        rpc_exchange: Optional[str] = None,
        events_exchange: Optional[str] = None,
        transport_settings: Optional[Dict[str, Any]] = None,
    ):
        """RMQAPI: A NuropbInterface implementation that uses RabbitMQ as the underlying transport."""
        parts = amqp_url.split("/")
        vhost = amqp_url.split("/")[-1]
        if len(parts) < 4:
            raise ValueError("Invalid amqp_url, missing vhost")
        self._mesh_name = vhost

        """ If a service_name is not provided, then the service is a client only and will not be able 
        to register for messages on service exchanges: rpc and events.
        """
        self._instance_id = instance_id if instance_id is not None else uuid4().hex

        if service_name is None:
            """Configure for client only mode"""
            self._client_only = True
            self._connection_name = f"{vhost}-client-{instance_id}"
            service_name = f"{vhost}-client"
            self._encryptor = Encryptor()
        else:
            """Configure for service mode"""
            self._client_only = False
            self._connection_name = f"{vhost}-{service_name}-{instance_id}"
            self._encryptor = Encryptor(
                service_name=service_name,
                private_key=getattr(service_instance, "_private_key", None),
            )

        self._service_name = service_name
        """ Is also a label for the api whether in client or service mode.
        """

        self._service_discovery = {}
        """ A dictionary of service_name: service mesh discovered service_info
        """
        self._service_public_keys = {}
        """ A dictionary of service_name: public_key for when encryption is required
        """

        if not self._client_only and service_instance is None:
            raise ValueError(
                "A service instance must be provided when starting in service mode"
            )  # pragma: no cover
        self._service_instance = service_instance
        """ the class instance that will be shared on the service mesh
        """

        transport_settings = {} if transport_settings is None else transport_settings
        if self._client_only:
            transport_settings["client_only"] = True

        self._response_futures = {}

        default_ttl = transport_settings.get("default_ttl", None)
        self._default_ttl = 60 * 60 * 1000 if default_ttl is None else default_ttl
        """ default time to live or timeout service mesh interaction
        """

        self._api_connected = False

        if not self._client_only and self._service_instance is None:
            logger.warning(
                "No service instance provided, service will not be able to handle requests"
            )  # pragma: no cover

        transport_settings.update(
            {
                "service_name": self._service_name,
                "instance_id": self._instance_id,
                "amqp_url": amqp_url,
                "message_callback": self.receive_transport_message,
                "rpc_exchange": rpc_exchange,
                "events_exchange": events_exchange,
                "encryptor": self._encryptor,
            }
        )
        self._transport = RMQTransport(**transport_settings)
        self._rpc_exchange = self._transport.rpc_exchange
        self._events_exchange = self._transport.events_exchange

    @property
    def service_name(self) -> str:
        """service_name: returns the service_name of the underlying transport"""
        return self._transport.service_name

    @property
    def is_leader(self) -> bool:
        return self._transport.is_leader

    @property
    def client_only(self) -> bool:
        """client_only: returns the client_only status of the underlying transport"""
        return self._client_only

    @property
    def connected(self) -> bool:
        """connected: returns the connection status of the underlying transport
        :return: bool
        """
        return self._transport.connected

    @property
    def transport(self) -> RMQTransport:
        """transport: returns the underlying transport
        :return: RMQTransport
        """
        return self._transport

    async def connect(self) -> None:
        """connect: connects to the underlying transport
        :return: None
        """
        if not self.connected:
            await self._transport.start()
        else:
            logger.warning("Already connected")

    async def disconnect(self) -> None:
        """disconnect: disconnects from the underlying transport
        :return: None
        """
        await self._transport.stop()

    def receive_transport_message(
        self,
        service_message: TransportServicePayload,
        message_complete_callback: MessageCompleteFunction,
        metadata: Dict[str, Any],
    ) -> None:
        """receive_transport_message: handles a messages received from the transport layer. Both
         incoming service messages and response messages pass through this method.

        :return: None
        """
        if service_message["nuropb_type"] == "response":
            response_payload: ResponsePayloadDict = cast(
                ResponsePayloadDict, service_message["nuropb_payload"]
            )
            logger.debug(
                f"Received response.  "
                f"trace_id: {service_message['trace_id']} "
                f"correlation_id: {service_message['correlation_id']}"
            )
            if service_message["correlation_id"] not in self._response_futures:
                logger.warning(
                    f"Received an unpaired response, ignoring "
                    f"correlation_id: {service_message['correlation_id']}"
                )
                return

            # Setting the result on this future will complete "await api.request(...)"
            try:
                response_future = self._response_futures.pop(
                    service_message["correlation_id"]
                )
                response_future.set_result(response_payload)
            except Exception as error:
                logger.exception(
                    f"Error completing response future for trace_id: {service_message['trace_id']} "
                    f"correlation_id: {service_message['correlation_id']} "
                    f"error: {error}"
                )

            return

        """ The logic below is only relevant for incoming service messages
        """
        if self._service_instance is None:
            error_description = f"No service instance configured to handle the {service_message['nuropb_type']} instruction"
            logger.warning(error_description)
            response = NuropbHandlingError(
                description=error_description,
                payload=service_message["nuropb_payload"],
            )
            handle_execution_result(service_message, response, MessageCompleteFunction)
            return

        if service_message["nuropb_type"] in ("request", "command", "event"):
            logger.debug(f"Received {service_message['nuropb_type']}")
            execute_request(
                self._service_instance, service_message, message_complete_callback
            )

        else:
            logger.warning(
                "Received an unsupported message type: %s",
                service_message["nuropb_type"],
            )

    async def request(
        self,
        service: str,
        method: str,
        params: Dict[str, Any],
        context: Dict[str, Any],
        ttl: Optional[int] = None,
        trace_id: Optional[str] = None,
        rpc_response: bool = True,
        encrypted: bool = False,
    ) -> Union[ResponsePayloadDict, Any]:
        """Makes a rpc request for a method on a service mesh service and waits until the response is
        received.

        :param service: str, The routing key on the rpc exchange to direct the request to the desired
            service request queue.
        :param method: str, the name of the api call / method on the service
        :param params: dict, The method input parameters
        :param context: dict
            The context of the request. This is used to pass information to the service manager
            and is not used by the transport. Example content includes:
                - user_id: str  # a unique user identifier or token of the user that made the request
                - correlation_id: str  # a unique identifier of the request used to correlate the response to the
                                       # request or trace the request over the network (e.g. uuid4 hex string)
                - service: str
                - method: str
        :param ttl: int optional
            expiry is the time in milliseconds that the message will be kept on the queue before being moved
            to the dead letter queue. If None, then the message expiry configured on the transport is used.
        :param trace_id: str optional
            an identifier to trace the request over the network (e.g. uuid4 hex string)
        :param rpc_response: bool optional
            if True (default), the actual response of the RPC call is returned and where there was an error,
            that is raised as an exception.
            Where rpc_response is a ResponsePayloadDict, is returned.
        :param encrypted: bool
            if True then the message will be encrypted in transit
        :return ResponsePayloadDict | Any: representing the response from the requested service with any
            exceptions raised
        """
        correlation_id = uuid4().hex
        ttl = self._default_ttl if ttl is None else ttl
        message: RequestPayloadDict = {
            "tag": "request",
            "service": service,
            "method": method,
            "params": params,
            "correlation_id": correlation_id,
            "context": context,
            "trace_id": trace_id,
        }
        response_future: ResultFutureResponsePayload = Future()
        self._response_futures[correlation_id] = response_future

        # mandatory means that if it doesn't get routed to a queue then it will be returned vi self._on_message_returned
        logger.debug(
            f"Sending request message:\n"
            f"correlation_id: {correlation_id}\n"
            f"trace_id: {trace_id}\n"
            f"service: {service}\n"
            f"method: {method}\n"
        )
        try:
            self._transport.send_message(
                payload=message,
                expiry=ttl,
                priority=None,
                encoding="json",
                encrypted=encrypted,
            )
        except Exception as err:
            if rpc_response is False:
                return {
                    "tag": "response",
                    "context": context,
                    "correlation_id": correlation_id,
                    "trace_id": trace_id,
                    "result": None,
                    "error": {
                        "error": f"{type(err).__name__}",
                        "description": f"Error sending request message: {err}",
                    },
                }
            else:
                raise err

        try:
            response = await response_future
            if rpc_response is True and response["error"] is not None:
                raise NuropbMessageError(
                    description=response["error"]["description"],
                    payload=response,
                )
            elif rpc_response is True:
                return response["result"]
            else:
                return response
        except BaseException as err:
            if rpc_response is True:
                raise err
            else:
                if not isinstance(err, NuropbException):
                    error = {
                        "error": f"{type(err).__name__}",
                        "description": f"Error waiting for response: {err}",
                    }
                else:
                    error = err.to_dict()
                return {
                    "tag": "response",
                    "context": context,
                    "correlation_id": correlation_id,
                    "trace_id": trace_id,
                    "result": None,
                    "error": error,
                }

    def command(
        self,
        service: str,
        method: str,
        params: Dict[str, Any],
        context: Dict[str, Any],
        ttl: Optional[int] = None,
        trace_id: Optional[str] = None,
        encrypted: bool = False,
    ) -> None:
        """command: sends a command to the target service. I.e. a targeted event. response is not expected
        and ignored.

        :param service: the service name
        :param method: the method name
        :param params: the method arguments, these must be easily serializable to JSON
        :param context: additional information that represent the context in which the request is executed.
                        The must be easily serializable to JSON.
        :param ttl: the time to live of the request in milliseconds. After this time and dependent on the
                    underlying transport, it will not be consumed by the target
                    or
                    assumed by the requester to have failed with an undetermined state.
        :param trace_id: an identifier to trace the request over the network (e.g. uuid4 hex string)
        :param encrypted: bool, if True then the message will be encrypted in transit
        :return: None
        """
        correlation_id = uuid4().hex
        ttl = self._default_ttl if ttl is None else ttl
        message: CommandPayloadDict = {
            "tag": "command",
            "service": service,
            "method": method,
            "params": params,
            "correlation_id": correlation_id,
            "context": context,
            "trace_id": trace_id,
        }

        # mandatory means that if it doesn't get routed to a queue then it will be returned vi self._on_message_returned
        logger.debug(
            f"Sending command message:\n"
            f"correlation_id: {correlation_id}\n"
            f"trace_id: {trace_id}\n"
            f"service: {service}\n"
            f"method: {method}\n"
        )
        self._transport.send_message(
            payload=message,
            expiry=ttl,
            priority=None,
            encoding="json",
            encrypted=encrypted,
        )

    def publish_event(
        self,
        topic: str,
        event: Dict[str, Any],
        context: Dict[str, Any],
        trace_id: Optional[str] = None,
        encrypted: bool = False,
    ) -> None:
        """Broadcasts an event with the given topic.

        :param topic: str, The routing key on the events exchange
        :param event: json-encodable Python Dict.
        :param context: dict, The context around gent generation, example content includes:
                - user_id: str  # a unique user identifier or token of the user that made the request
                - correlation_id: str  # a unique identifier of the request used to correlate the response
                                       # to the request
                                       # or trace the request over the network (e.g. an uuid4 hex string)
                - service: str
                - method: str
        :param trace_id: str optional
            an identifier to trace the request over the network (e.g. an uuid4 hex string)
        :param encrypted: bool, if True then the message will be encrypted in transit
        """
        correlation_id = uuid4().hex
        message: EventPayloadDict = {
            "tag": "event",
            "topic": topic,
            "event": event,
            "context": context,
            "trace_id": trace_id,
            "correlation_id": correlation_id,
            "target": None,
        }
        logger.debug(
            "Sending event message: (%s - %s)",
            correlation_id,
            trace_id,
        )
        self._transport.send_message(
            payload=message,
            priority=None,
            encoding="json",
            encrypted=encrypted,
        )

    async def describe_service(
        self, service_name: str, refresh: bool = False
    ) -> Dict[str, Any] | None:
        """describe_service: returns the service information for the given service_name,
        if it is not already cached or refresh is try then the service discovery is queried directly.

        :param service_name: str
        :param refresh: bool
        :return: dict
        """
        if service_name in self._service_discovery or refresh:
            return self._service_discovery[service_name]

        service_info = await self.request(
            service=service_name,
            method="nuropb_describe",
            params={},
            context={},
            ttl=60 * 1000,  # 1 minute
            trace_id=uuid4().hex,
        )
        if not isinstance(service_info, dict):
            raise ValueError(
                f"Invalid service_info returned for service {service_name}"
            )
        else:
            self._service_discovery[service_name] = service_info
            try:
                text_public_key = service_info.get("public_key", None)
                if text_public_key:
                    public_key = serialization.load_pem_public_key(
                        data=text_public_key.encode("ascii"),
                        backend=default_backend(),
                    )
                    self._encryptor.add_service_public_key(
                        service_name=service_name, public_key=public_key
                    )
            except Exception as err:
                logger.error(f"error loading the public key for {service_name}: {err}")
            finally:
                return service_info

    async def requires_encryption(self, service_name: str, method_name: str) -> bool:
        """requires_encryption: Queries the service discovery information for the service_name
        and returns True if encryption is required else False.
        none of encryption is not required.

        :param service_name: str
        :param method_name: str
        :return: bool
        """
        service_info = await self.describe_service(service_name)
        method_info = service_info["methods"].get(method_name, None)
        if method_info is None:
            raise ValueError(
                f"Method {method_name} not found on service {service_name}"
            )
        return method_info.get("requires_encryption", False)
